import sys
import warnings
import os
from pathlib import Path

import bokeh.palettes as bp
from bokeh.models import HoverTool, ColumnDataSource
import matplotlib.patches as mpatches
import numpy as np
import toml
import pandas as pd
import xlrd

import ross
from ross.element import Element
from ross.materials import Material
from ross.materials import steel

__all__ = ["ShaftElement"]

bokeh_colors = bp.RdGy[11]


class ShaftElement(Element):
    r"""A shaft element.
    This class will create a shaft element that may take into
    account shear, rotary inertia an gyroscopic effects.
    The matrices will be defined considering the following local
    coordinate vector:
    .. math:: [x_0, y_0, \alpha_0, \beta_0, x_1, y_1, \alpha_1, \beta_1]^T
    Where :math:`\alpha_0` and :math:`\alpha_1` are the bending on the yz plane and
    :math:`\beta_0` and :math:`\beta_1` are the bending on the xz plane.
    Parameters
    ----------
    L : float
        Element length.
    i_d : float
        Inner diameter of the element.
    o_d : float
        Outer diameter of the element.
    material : ross.material
        Shaft material.
    n : int, optional
        Element number (coincident with it's first node).
        If not given, it will be set when the rotor is assembled
        according to the element's position in the list supplied to
        the rotor constructor.
    axial_force : float, optional
        Axial force.
    torque : float, optional
        Torque.
    shear_effects : bool, optional
        Determine if shear effects are taken into account.
        Default is True.
    rotary_inertia : bool, optional
        Determine if rotary_inertia effects are taken into account.
        Default is True.
    gyroscopic : bool, optional
        Determine if gyroscopic effects are taken into account.
        Default is True.
    shear_method_calc : string, optional
        Determines which shear calculation method the user will adopt
        Default is 'cowper'
    Returns
    -------
    Attributes
    ----------
    Poisson : float
        Poisson coefficient for the element.
    A : float
        Element section area.
    Ie : float
        Ie is the second moment of area of the cross section about
        the neutral plane Ie = pi*r**2/4
    phi : float
        Constant that is used according to [1]_ to consider rotary
        inertia and shear effects. If these are not considered phi=0.
    References
    ----------
    .. [1] 'Dynamics of Rotating Machinery' by MI Friswell, JET Penny, SD Garvey
       & AW Lees, published by Cambridge University Press, 2010 pp. 158-166.
    Examples
    --------
    >>> from ross.materials import steel
    >>> le = 0.25
    >>> i_d = 0
    >>> o_d = 0.05
    >>> Euler_Bernoulli_Element = ShaftElement(le, i_d, o_d, steel,
    ...                                        shear_effects=False, rotary_inertia=False)
    >>> Euler_Bernoulli_Element.phi
    0
    >>> Timoshenko_Element = ShaftElement(le, i_d, o_d, steel,
    ...                                   rotary_inertia=True,
    ...                                   shear_effects=True)
    >>> Timoshenko_Element.phi
    0.08795566502463055
    """

    def __init__(
        self,
        L,
        i_d,
        o_d,
        material,
        n=None,
        axial_force=0,
        torque=0,
        shear_effects=True,
        rotary_inertia=True,
        gyroscopic=True,
        shear_method_calc="cowper",
    ):

        if type(material) is str:
            os.chdir(Path(os.path.dirname(ross.__file__)))
            self.material = Material.use_material(material)
        else:
            self.material = material

        self.shear_effects = shear_effects
        self.rotary_inertia = rotary_inertia
        self.gyroscopic = gyroscopic
        self.axial_force = axial_force
        self.torque = torque
        self._n = n
        self.n_l = n
        self.n_r = None
        if n is not None:
            self.n_r = n + 1

        self.shear_method_calc = shear_method_calc

        self.L = float(L)

        # diameters
        self.i_d = float(i_d)
        self.o_d = float(o_d)
        self.i_d_l = float(i_d)
        self.o_d_l = float(o_d)
        self.i_d_r = float(i_d)
        self.o_d_r = float(o_d)
        self.color = self.material.color

        self.A = np.pi * (o_d ** 2 - i_d ** 2) / 4
        self.volume = self.A * self.L
        self.m = self.material.rho * self.volume
        #  Ie is the second moment of area of the cross section about
        #  the neutral plane Ie = pi*r**2/4
        self.Ie = np.pi * (o_d ** 4 - i_d ** 4) / 64
        phi = 0

        # Slenderness ratio of beam elements (G*A*L^2) / (E*I)
        sld = (self.material.G_s * self.A * self.L ** 2) / (self.material.E * self.Ie)
        self.slenderness_ratio = sld

        # picking a method to calculate the shear coefficient
        # List of avaible methods:
        # hutchinson - kappa as per Hutchinson (2001)
        # cowper - kappa as per Cowper (1996)
        if shear_effects:
            r = i_d / o_d
            r2 = r * r
            r12 = (1 + r2) ** 2
            if shear_method_calc == "hutchinson":
                # Shear coefficient (phi)
                # kappa as per Hutchinson (2001)
                # fmt: off
                kappa = 6*r12*((1+self.material.Poisson)/
                        ((r12*(7 + 12*self.material.Poisson + 4*self.material.Poisson**2) +
                        4*r2*(5 + 6*self.material.Poisson + 2*self.material.Poisson**2))))
                # fmt: on
            elif shear_method_calc == "cowper":
                # kappa as per Cowper (1996)
                # fmt: off
                kappa = 6 * r12 * (
                    (1 + self.material.Poisson)
                    / (r12 * (7 + 6 * self.material.Poisson) + r2 * (20 + 12 * self.material.Poisson))
                )
                # fmt: on
            else:
                raise Warning(
                    "This method of calculating shear coefficients is not implemented. See guide for futher informations."
                )

            # fmt: off
            phi = 12 * self.material.E * self.Ie / (self.material.G_s * kappa * self.A * L ** 2)
            # fmt: on

        self.phi = phi

    def __eq__(self, other):
        if self.__dict__ == other.__dict__:
            return True
        else:
            return False

    def save(self, file_name):
        data = self.load_data(file_name)
        data["ShaftElement"][str(self.n)] = {
            "L": self.L,
            "i_d": self.i_d,
            "o_d": self.o_d,
            "material": self.material.name,
            "n": self.n,
            "axial_force": self.axial_force,
            "torque": self.torque,
            "shear_effects": self.shear_effects,
            "rotary_inertia": self.rotary_inertia,
            "gyroscopic": self.gyroscopic,
            "shear_method_calc": self.shear_method_calc,
        }
        self.dump_data(data, file_name)

    @staticmethod
    def load(file_name="ShaftElement"):
        shaft_elements = []
        with open("ShaftElement.toml", "r") as f:
            shaft_elements_dict = toml.load(f)
            for element in shaft_elements_dict["ShaftElement"]:
                shaft_elements.append(
                    ShaftElement(**shaft_elements_dict["ShaftElement"][element])
                )
        return shaft_elements

    @classmethod
    def from_table(cls, file, sheet="Simple"):
        """Instantiate one or more shafts using inputs from a table, either excel or csv.
        Parameters
        ----------
        file: str
            Path to the file containing the shaft parameters.
        sheet: str, optional
            Describes the kind of sheet the function should expect:
                Simple: The input table should contain a header with column names equal
                to parameter names in the ShaftElement class, except for
                shear_effects, rotary_inertia, gyroscopic, and shear_method_calc.
                Model: The sheet must follow the expected format. The function will look
                for the parameters according to this format, in the positions they are supposed
                to be found. Headers should be in rows 4 and 20, just before parameters.
        Returns
        -------
        shaft : list
            A list of shaft objects.
        """
        if sheet == "Simple":
            try:
                df = pd.read_excel(file)
            except FileNotFoundError:
                sys.exit(file + " not found.")
            except xlrd.biffh.XLRDError:
                df = pd.read_csv(file)
            try:
                for index, row in df.iterrows():
                    for i in range(0, row.size):
                        if pd.isna(row[i]):
                            warnings.warn(
                                "NaN found in row "
                                + str(index)
                                + " column "
                                + str(i)
                                + ".\n"
                                "It will be replaced with zero."
                            )
                            row[i] = 0
                list_of_shafts = []
                for i, row in df.iterrows():
                    shear_effects = True
                    rotary_inertia = True
                    gyroscopic = True
                    shear_method_calc = "cowper"
                    try:
                        shear_effects = bool(row["shear_effects"])
                    except KeyError:
                        pass
                    try:
                        rotary_inertia = bool(row["rotary_inertia"])
                    except KeyError:
                        pass
                    try:
                        gyroscopic = bool(row["gyroscopic"])
                    except KeyError:
                        pass
                    try:
                        shear_method_calc = row["shear_method_calc"]
                    except KeyError:
                        pass
                    list_of_shafts.append(
                        cls(
                            row.L,
                            row.i_d,
                            row.o_d,
                            Material.use_material(row.material),
                            n=row.n,
                            axial_force=row.axial_force,
                            torque=row.torque,
                            shear_effects=shear_effects,
                            rotary_inertia=rotary_inertia,
                            gyroscopic=gyroscopic,
                            shear_method_calc=shear_method_calc,
                        )
                    )
                return list_of_shafts
            except KeyError:
                sys.exit(
                    "One or more column names did not match the expected. "
                    "Make sure the table header contains the parameters for the "
                    "ShaftElement class. Also, make sure you have a material "
                    "with the given name."
                )
        elif sheet == "Model":
            try:
                df1 = pd.read_excel(file, header=3, nrows=10)
                df2 = pd.read_excel(file, header=19)
                df_unit = pd.read_excel(file, header=16, nrows=2)
            except FileNotFoundError:
                sys.exit(file + " not found.")
            except xlrd.biffh.XLRDError:
                df1 = pd.read_csv(file, header=3, nrows=10)
                df2 = pd.read_csv(file, header=19)
                df_unit = pd.read_csv(file, header=16, nrows=2)
            convert_to_metric = False
            if df_unit["Length"][1] != "meters":
                convert_to_metric = True
            material_name = []
            material_rho = []
            material_e = []
            material_g_s = []
            new_materials = {}
            for index, row in df1.iterrows():
                if not pd.isna(row["matno"]):
                    material_name.append(int(row["matno"]))
                    material_rho.append(row["rhoa"])
                    material_e.append(row["ea"])
                    material_g_s.append(row["ga"])
            if convert_to_metric:
                for i in range(0, len(material_name)):
                    material_rho[i] = material_rho[i] * 27679.904
                    material_e[i] = material_e[i] * 6894.757
                    material_g_s[i] = material_g_s[i] * 6894.757
            for i in range(0, len(material_name)):
                new_material = Material(
                    name="shaft_mat_" + str(material_name[i]),
                    rho=material_rho[i],
                    E=material_e[i],
                    G_s=material_g_s[i],
                )
                new_materials["shaft_mat_" + str(material_name[i])] = new_material
            shaft_l = []
            shaft_i_d = []
            shaft_o_d = []
            shaft_material = []
            shaft_n = []
            shaft_axial_force = []
            for index, row in df2.iterrows():
                shaft_l.append(row["length"])
                shaft_i_d.append(row["id_Left"])
                shaft_o_d.append(row["od_Left"])
                shaft_material.append(int(row["matnum"]))
                shaft_n.append(row["elemnum"] - 1)
                shaft_axial_force.append(row["axial"])
            if convert_to_metric:
                for i in range(0, len(shaft_n)):
                    shaft_l[i] = shaft_l[i] * 0.0254
                    shaft_i_d[i] = shaft_i_d[i] * 0.0254
                    shaft_o_d[i] = shaft_o_d[i] * 0.0254
                    shaft_axial_force[i] = shaft_axial_force[i] * 4.448_221_61
            list_of_shafts = []
            for i in range(0, len(shaft_n)):
                list_of_shafts.append(
                    cls(
                        shaft_l[i],
                        shaft_i_d[i],
                        shaft_o_d[i],
                        new_materials["shaft_mat_" + str(shaft_material[i])],
                        n=shaft_n[i],
                        axial_force=shaft_axial_force[i],
                    )
                )
            return list_of_shafts
        else:
            sys.exit(
                "A valid choice must be given for the parameter 'sheet'. Either 'Simple' or 'Model' "
                "were expected. It was given " + sheet + "."
            )

    @property
    def n(self):
        return self._n

    @n.setter
    def n(self, value):
        self._n = value
        self.n_l = value
        if value is not None:
            self.n_r = value + 1

    def dof_mapping(self):
        return dict(x0=0, y0=1, alpha0=2, beta0=3, x1=4, y1=5, alpha1=6, beta1=7)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(L={self.L:{0}.{5}}, i_d={self.i_d:{0}.{5}}, "
            f"o_d={self.o_d:{0}.{5}}, material={self.material.name!r}, "
            f"n={self.n})"
        )

    def __str__(self):
        return (
            f"\nElem. N:    {self.n}"
            f"\nLenght:     {self.L:{10}.{5}}"
            f"\nInt. Diam.: {self.i_d:{10}.{5}}"
            f"\nOut. Diam.: {self.o_d:{10}.{5}}"
            f'\n{35*"-"}'
            f"\n{self.material}"
            f"\n"
        )

    def M(self):
        r"""Mass matrix for an instance of a shaft element.
        Returns
        -------
        Mass matrix for the shaft element.
        Examples
        --------
        >>> Timoshenko_Element = ShaftElement(0.25, 0, 0.05, steel,
        ...                                  rotary_inertia=True,
        ...                                  shear_effects=True)
        >>> Timoshenko_Element.M()[:4, :4]
        array([[ 1.42050794,  0.        ,  0.        ,  0.04931719],
               [ 0.        ,  1.42050794, -0.04931719,  0.        ],
               [ 0.        , -0.04931719,  0.00231392,  0.        ],
               [ 0.04931719,  0.        ,  0.        ,  0.00231392]])
        """
        phi = self.phi
        L = self.L

        m01 = 312 + 588 * phi + 280 * phi ** 2
        m02 = (44 + 77 * phi + 35 * phi ** 2) * L
        m03 = 108 + 252 * phi + 140 * phi ** 2
        m04 = -(26 + 63 * phi + 35 * phi ** 2) * L
        m05 = (8 + 14 * phi + 7 * phi ** 2) * L ** 2
        m06 = -(6 + 14 * phi + 7 * phi ** 2) * L ** 2
        # fmt: off
        M = np.array([[m01,    0,    0,  m02,  m03,    0,    0,  m04],
                      [  0,  m01, -m02,    0,    0,  m03, -m04,    0],
                      [  0, -m02,  m05,    0,    0,  m04,  m06,    0],
                      [m02,    0,    0,  m05, -m04,    0,    0,  m06],
                      [m03,    0,    0, -m04,  m01,    0,    0, -m02],
                      [  0,  m03,  m04,    0,    0,  m01,  m02,    0],
                      [  0, -m04,  m06,    0,    0,  m02,  m05,    0],
                      [m04,    0,    0,  m06, -m02,    0,    0,  m05]])
        # fmt: on
        M = self.material.rho * self.A * self.L * M / (840 * (1 + phi) ** 2)

        if self.rotary_inertia:
            ms1 = 36
            ms2 = (3 - 15 * phi) * L
            ms3 = (4 + 5 * phi + 10 * phi ** 2) * L ** 2
            ms4 = (-1 - 5 * phi + 5 * phi ** 2) * L ** 2
            # fmt: off
            Ms = np.array([[ ms1,    0,    0,  ms2, -ms1,    0,    0,  ms2],
                           [   0,  ms1, -ms2,    0,    0, -ms1, -ms2,    0],
                           [   0, -ms2,  ms3,    0,    0,  ms2,  ms4,    0],
                           [ ms2,    0,    0,  ms3, -ms2,    0,    0,  ms4],
                           [-ms1,    0,    0, -ms2,  ms1,    0,    0, -ms2],
                           [   0, -ms1,  ms2,    0,    0,  ms1,  ms2,    0],
                           [   0, -ms2,  ms4,    0,    0,  ms2,  ms3,    0],
                           [ ms2,    0,    0,  ms4, -ms2,    0,    0,  ms3]])
            # fmt: on
            Ms = self.material.rho * self.Ie * Ms / (30 * L * (1 + phi) ** 2)
            M = M + Ms

        return M

    def K(self):
        r"""Stiffness matrix for an instance of a shaft element.
        Returns
        -------
        Stiffness matrix for the shaft element.
        Examples
        --------
        >>> from ross.materials import steel
        >>> Timoshenko_Element = ShaftElement(0.25, 0, 0.05, steel,
        ...                                  rotary_inertia=True,
        ...                                  shear_effects=True)
        >>> Timoshenko_Element.K()[:4, :4]/1e6
        array([[45.69644273,  0.        ,  0.        ,  5.71205534],
               [ 0.        , 45.69644273, -5.71205534,  0.        ],
               [ 0.        , -5.71205534,  0.97294287,  0.        ],
               [ 5.71205534,  0.        ,  0.        ,  0.97294287]])
        """
        phi = self.phi
        L = self.L
        # fmt: off
        K = np.array([
            [12,     0,            0,          6*L,  -12,     0,            0,          6*L],
            [0,     12,         -6*L,            0,    0,   -12,         -6*L,            0],
            [0,   -6*L, (4+phi)*L**2,            0,    0,   6*L, (2-phi)*L**2,            0],
            [6*L,    0,            0, (4+phi)*L**2, -6*L,     0,            0, (2-phi)*L**2],
            [-12,    0,            0,         -6*L,   12,     0,            0,         -6*L],
            [0,    -12,          6*L,            0,    0,    12,          6*L,            0],
            [0,   -6*L, (2-phi)*L**2,            0,    0,   6*L, (4+phi)*L**2,            0],
            [6*L,    0,            0, (2-phi)*L**2, -6*L,     0,            0, (4+phi)*L**2]
        ])
        # fmt: on
        K = self.material.E * self.Ie * K / ((1 + phi) * L ** 3)

        return K

    def C(self):
        """Stiffness matrix for an instance of a shaft element.

        Returns
        -------
        C : np.array
           Damping matrix for the shaft element.

        Examples
        --------
        """
        C = np.zeros((8, 8))

        return C

    def G(self):
        """Gyroscopic matrix for an instance of a shaft element.
        Returns
        -------
        Gyroscopic matrix for the shaft element.
        Examples
        --------
        >>> from ross.materials import steel
        >>> # Timoshenko is the default shaft element
        >>> Timoshenko_Element = ShaftElement(0.25, 0, 0.05, steel)
        >>> Timoshenko_Element.G()[:4, :4]
        array([[-0.        ,  0.01943344, -0.00022681, -0.        ],
               [-0.01943344, -0.        , -0.        , -0.00022681],
               [ 0.00022681, -0.        , -0.        ,  0.0001524 ],
               [-0.        ,  0.00022681, -0.0001524 , -0.        ]])
        """
        phi = self.phi
        L = self.L

        G = np.zeros((8, 8))

        if self.gyroscopic:
            g1 = 36
            g2 = (3 - 15 * phi) * L
            g3 = (4 + 5 * phi + 10 * phi ** 2) * L ** 2
            g4 = (-1 - 5 * phi + 5 * phi ** 2) * L ** 2
            # fmt: off
            G = np.array([[  0, -g1,  g2,   0,   0,  g1,  g2,   0],
                          [ g1,   0,   0,  g2, -g1,   0,   0,  g2],
                          [-g2,   0,   0, -g3,  g2,   0,   0, -g4],
                          [  0, -g2,  g3,   0,   0,  g2,  g4,   0],
                          [  0,  g1, -g2,   0,   0, -g1, -g2,   0],
                          [-g1,   0,   0, -g2,  g1,   0,   0, -g2],
                          [-g2,   0,   0, -g4,  g2,   0,   0, -g3],
                          [  0, -g2,  g4,   0,   0,  g2,  g3,   0]])
            # fmt: on
            G = -self.material.rho * self.Ie * G / (15 * L * (1 + phi) ** 2)

        return G

    def patch(self, position, SR, ax):
        """Shaft element patch.
        Patch that will be used to draw the shaft element.
        Parameters
        ----------
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        bk_ax : bokeh plotting axes, optional
            Axes in which the plot will be drawn.
        position : float
            Position in which the patch will be drawn.
        Returns
        -------
        """
        position_u = [position, self.i_d / 2]  # upper
        position_l = [position, -self.o_d / 2]  # lower
        width = self.L
        height = self.o_d / 2 - self.i_d / 2
        if self.n in SR:
            mpl_color = "yellow"
            bk_color = "yellow"
            legend = "Shaft - Slenderness Ratio < 1.6"
        else:
            mpl_color = self.color
            bk_color = bokeh_colors[2]
            legend = "Shaft"

        #  matplotlib - plot the upper half of the shaft
        ax.add_patch(
            mpatches.Rectangle(
                position_u,
                width,
                height,
                linestyle="--",
                linewidth=0.5,
                ec="k",
                fc=mpl_color,
                alpha=0.8,
                label=legend,
            )
        )
        #  matplotlib - plot the lower half of the shaft
        ax.add_patch(
            mpatches.Rectangle(
                position_l,
                width,
                height,
                linestyle="--",
                linewidth=0.5,
                ec="k",
                fc=mpl_color,
                alpha=0.8,
            )
        )

    def bokeh_patch(self, position, SR, bk_ax):
        """Shaft element patch.
        Patch that will be used to draw the shaft element.
        Parameters
        ----------
        bk_ax : bokeh plotting axes, optional
            Axes in which the plot will be drawn.
        position : float
            Position in which the patch will be drawn.
        Returns
        -------
        """

        if self.n in SR:
            bk_color = "yellow"
            legend = "Shaft - Slenderness Ratio < 1.6"
        else:
            bk_color = bokeh_colors[2]
            legend = "Shaft"

        source_u = ColumnDataSource(
            dict(
                top=[self.o_d / 2],
                bottom=[self.i_d / 2],
                left=[position],
                right=[position + self.L],
                elnum=[self.n],
                out_d=[self.o_d],
                in_d=[self.i_d],
                length=[self.L],
                mat=[self.material.name],
            )
        )

        source_l = ColumnDataSource(
            dict(
                top=[-self.o_d / 2],
                bottom=[-self.i_d / 2],
                left=[position],
                right=[position + self.L],
                elnum=[self.n],
                out_d=[self.o_d],
                in_d=[self.i_d],
                length=[self.L],
                mat=[self.material.name],
            )
        )

        # bokeh plot - plot the shaft
        bk_ax.quad(
            top="top",
            bottom="bottom",
            left="left",
            right="right",
            source=source_u,
            line_color=bokeh_colors[0],
            line_width=1,
            fill_alpha=0.5,
            fill_color=bk_color,
            legend=legend,
            name="u_shaft",
        )
        bk_ax.quad(
            top="top",
            bottom="bottom",
            left="left",
            right="right",
            source=source_l,
            line_color=bokeh_colors[0],
            line_width=1,
            fill_alpha=0.5,
            fill_color=bk_color,
            name="l_shaft",
        )
        hover = HoverTool(names=["u_shaft", "l_shaft"])
        hover.tooltips = [
            ("Element Number :", "@elnum"),
            ("Outer Diameter :", "@out_d"),
            ("Internal Diameter :", "@in_d"),
            ("Element Length :", "@length"),
            ("Material :", "@mat"),
        ]
        hover.mode = "mouse"

        if len(bk_ax.hover) == 0:
            bk_ax.add_tools(hover)

    @classmethod
    def section(
        cls,
        L,
        ne,
        si_d,
        so_d,
        material,
        n=None,
        shear_effects=True,
        rotary_inertia=True,
        gyroscopic=True,
    ):
        """Shaft section constructor.
        This method will create a shaft section with length 'L'
        divided into 'ne' elements.
        Parameters
        ----------
        i_d : float
            Inner diameter of the section.
        o_d : float
            Outer diameter of the section.
        E : float
            Young's modulus.
        G_s : float
            Shear modulus.
        material : ross.material
            Shaft material.
        n : int, optional
            Element number (coincident with it's first node).
            If not given, it will be set when the rotor is assembled
            according to the element's position in the list supplied to
            the rotor constructor.
        axial_force : float
            Axial force.
        torque : float
            Torque.
        shear_effects : bool
            Determine if shear effects are taken into account.
            Default is False.
        rotary_inertia : bool
            Determine if rotary_inertia effects are taken into account.
            Default is False.
        gyroscopic : bool
            Determine if gyroscopic effects are taken into account.
            Default is False.
        Returns
        -------
        elements: list
            List with the 'ne' shaft elements.
        Examples
        --------
        >>> # shaft material
        >>> from ross.materials import steel
        >>> # shaft inner and outer diameters
        >>> si_d = 0
        >>> so_d = 0.01585
        >>> sec = ShaftElement.section(247.65e-3, 4, 0, 15.8e-3, steel)
        >>> len(sec)
        4
        >>> sec[0].i_d
        0.0
        """

        le = L / ne

        elements = [
            cls(le, si_d, so_d, material, n, shear_effects, rotary_inertia, gyroscopic)
            for _ in range(ne)
        ]

        return elements



import sys
import warnings

import bokeh.palettes as bp
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.interpolate as interpolate
import xlrd

from ross.element import Element
from ross.fluid_flow import fluid_flow as flow

__all__ = ["BearingElement", "SealElement"]
bokeh_colors = bp.RdGy[11]


class _Coefficient:
    def __init__(self, coefficient, w=None, interpolated=None):
        if isinstance(coefficient, (int, float)):
            if w is not None and type(w) != float:
                coefficient = [coefficient for _ in range(len(w))]
            else:
                coefficient = [coefficient]

        self.coefficient = coefficient
        self.w = w

        if len(self.coefficient) > 1:
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    self.interpolated = interpolate.UnivariateSpline(
                        self.w, self.coefficient
                    )
            #  dfitpack.error is not exposed by scipy
            #  so a bare except is used
            except:
                raise ValueError(
                    "Arguments (coefficients and w)" " must have the same dimension"
                )
        else:
            self.interpolated = lambda x: np.array(self.coefficient[0])

    def __eq__(self, other):
        if np.allclose(self.__dict__["coefficient"],other.__dict__["coefficient"]):
            return True
        else:
            return False

    def __repr__(self):
        coef = []
        for i in self.coefficient:
            coef.append("{:.2e}".format(i))
        return f"{coef}"

    def plot(self, ax=None, **kwargs):
        if ax is None:
            ax = plt.gca()

        w_range = np.linspace(min(self.w), max(self.w), 30)

        ax.plot(w_range, self.interpolated(w_range), **kwargs)
        ax.set_xlabel("Speed (rad/s)")
        ax.ticklabel_format(style="sci", scilimits=(0, 0), axis="y")

        return ax


class _Stiffness_Coefficient(_Coefficient):
    def plot(self, **kwargs):
        ax = super().plot(**kwargs)
        ax.set_ylabel("Stiffness ($N/m$)")

        return ax


class _Damping_Coefficient(_Coefficient):
    def plot(self, **kwargs):
        ax = super().plot(**kwargs)
        ax.set_ylabel("Damping ($Ns/m$)")

        return ax


class BearingElement(Element):
    """A bearing element.
    This class will create a bearing element.
    Parameters can be a constant value or speed dependent.
    For speed dependent parameters, each argument should be passed
    as an array and the correspondent speed values should also be
    passed as an array.
    Values for each parameter will be interpolated for the speed.
    Parameters
    ----------
    n: int
        Node which the bearing will be located in
    kxx: float, array
        Direct stiffness in the x direction.
    cxx: float, array
        Direct damping in the x direction.
    kyy: float, array, optional
        Direct stiffness in the y direction.
        (defaults to kxx)
    cyy: float, array, optional
        Direct damping in the y direction.
        (defaults to cxx)
    kxy: float, array, optional
        Cross coupled stiffness in the x direction.
        (defaults to 0)
    cxy: float, array, optional
        Cross coupled damping in the x direction.
        (defaults to 0)
    kyx: float, array, optional
        Cross coupled stiffness in the y direction.
        (defaults to 0)
    cyx: float, array, optional
        Cross coupled damping in the y direction.
        (defaults to 0)
    w: array, optional
        Array with the speeds (rad/s).
    Examples
    --------
    >>> # A bearing element located in the first rotor node, with these
    >>> # following stiffness and damping coefficients and speed range from
    >>> # 0 to 200 rad/s
    >>> import ross as rs
    >>> kxx = 1e6
    >>> kyy = 0.8e6
    >>> cxx = 2e2
    >>> cyy = 1.5e2
    >>> w = np.linspace(0, 200, 11)
    >>> bearing0 = rs.BearingElement(n=0, kxx=kxx, kyy=kyy, cxx=cxx, cyy=cyy, w=w)
    >>> bearing0.K(w) # doctest: +ELLIPSIS
    array([[[1000000., 1000000., ...
    >>> bearing0.C(w) # doctest: +ELLIPSIS
    array([[[200., 200., ...
    """

    def __init__(
        self, n, kxx, cxx, kyy=None, kxy=0, kyx=0, cyy=None, cxy=0, cyx=0, w=None
    ):

        args = ["kxx", "kyy", "kxy", "kyx", "cxx", "cyy", "cxy", "cyx"]

        # all args to coefficients
        args_dict = locals()
        coefficients = {}

        if kyy is None:
            args_dict["kyy"] = kxx
        if cyy is None:
            args_dict["cyy"] = cxx

        for arg in args:
            if arg[0] == "k":
                coefficients[arg] = _Stiffness_Coefficient(
                    coefficient=args_dict[arg], w=args_dict["w"]
                )
            else:
                coefficients[arg] = _Damping_Coefficient(args_dict[arg], args_dict["w"])

        coefficients_len = [len(v.coefficient) for v in coefficients.values()]

        if w is not None and type(w) != float:
            coefficients_len.append(len(args_dict["w"]))
            if len(set(coefficients_len)) > 1:
                raise ValueError(
                    "Arguments (coefficients and w)" " must have the same dimension"
                )
        else:
            for c in coefficients_len:
                if c != 1:
                    raise ValueError(
                        "Arguments (coefficients and w)" " must have the same dimension"
                    )

        for k, v in coefficients.items():
            setattr(self, k, v)

        self.n = n
        self.n_l = n
        self.n_r = n

        self.w = np.array(w, dtype=np.float64)
        self.color = "#355d7a"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(n={self.n},\n"
            f" kxx={self.kxx}, kxy={self.kxy},\n"
            f" kyx={self.kyx}, kyy={self.kxy},\n"
            f" cxx={self.cxx}, cxy={self.cxy},\n"
            f" cyx={self.cyx}, cyy={self.cyy},\n"
            f" w={self.w})"
        )

    def __eq__(self, other):
        compared_attributes = [
            "kxx",
            "kyy",
            "kxy",
            "kyx",
            "cxx",
            "cyy",
            "cxy",
            "cyx",
            "w",
        ]
        if isinstance(other, self.__class__):
            return all(
                (
                    np.array(getattr(self, attr)).all()
                    == np.array(getattr(other, attr)).all()
                    for attr in compared_attributes
                )
            )
        return False

    def save(self, file_name):
        data = self.load_data(file_name)
        if type(self.w) == np.ndarray:
            try:
                self.w[0]
                w = list(self.w)
            except IndexError:
                w = None
        data["BearingElement"][str(self.n)] = {
            "n": self.n,
            "kxx": self.kxx.coefficient,
            "cxx": self.cxx.coefficient,
            "kyy": self.kyy.coefficient,
            "kxy": self.kxy.coefficient,
            "kyx": self.kyx.coefficient,
            "cyy": self.cyy.coefficient,
            "cxy": self.cxy.coefficient,
            "cyx": self.cyx.coefficient,
            "w": w,
        }
        self.dump_data(data, file_name)

    @staticmethod
    def load(file_name="BearingElement"):
        bearing_elements = []
        bearing_elements_dict = BearingElement.load_data(
            file_name="BearingElement.toml"
        )
        for element in bearing_elements_dict["BearingElement"]:
            bearing = BearingElement(**bearing_elements_dict["BearingElement"][element])
            bearing.kxx.coefficient = bearing_elements_dict["BearingElement"][element][
                "kxx"
            ]

            bearing.kxy.coefficient = bearing_elements_dict["BearingElement"][element][
                "kxy"
            ]

            bearing.kyx.coefficient = bearing_elements_dict["BearingElement"][element][
                "kyx"
            ]

            bearing.kyy.coefficient = bearing_elements_dict["BearingElement"][element][
                "kyy"
            ]

            bearing.cxx.coefficient = bearing_elements_dict["BearingElement"][element][
                "cxx"
            ]

            bearing.cxy.coefficient = bearing_elements_dict["BearingElement"][element][
                "cxy"
            ]

            bearing.cyx.coefficient = bearing_elements_dict["BearingElement"][element][
                "cyx"
            ]

            bearing.cyy.coefficient = bearing_elements_dict["BearingElement"][element][
                "cyy"
            ]

            bearing_elements.append(bearing)
        return bearing_elements

    def dof_mapping(self):
        return dict(x0=0, y0=1)

    def M(self):
        M = np.zeros((2, 2))

        return M

    def K(self, w):
        kxx = self.kxx.interpolated(w)
        kyy = self.kyy.interpolated(w)
        kxy = self.kxy.interpolated(w)
        kyx = self.kyx.interpolated(w)

        K = np.array([[kxx, kxy], [kyx, kyy]])

        return K

    def C(self, w):
        cxx = self.cxx.interpolated(w)
        cyy = self.cyy.interpolated(w)
        cxy = self.cxy.interpolated(w)
        cyx = self.cyx.interpolated(w)

        C = np.array([[cxx, cxy], [cyx, cyy]])

        return C

    def G(self):
        G = np.zeros((2, 2))

        return G

    def patch(self, position, length, ax):
        """Bearing element patch.
        Patch that will be used to draw the bearing element.

        Parameters
        ----------
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        bk_ax : bokeh plotting axes, optional
            Axes in which the plot will be drawn.
        position : tuple
            Position (z, y) in which the patch will be drawn.
        length : float
            minimum length of shaft elements

        Returns
        -------
        """
        zpos, ypos = position
        le = length
        h = -0.5 * ypos  # height

        #  node (x pos), outer diam. (y pos)
        bearing_points = [
            [zpos, ypos],  # upper
            [zpos + h / 2, ypos - h],
            [zpos - h / 2, ypos - h],
            [zpos, ypos],
        ]
        ax.add_patch(mpatches.Polygon(bearing_points, color=self.color, picker=True))

    def bokeh_patch(self, position, length, bk_ax):
        """Bearing element patch.
        Patch that will be used to draw the bearing element.

        Parameters
        ----------
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        bk_ax : bokeh plotting axes, optional
            Axes in which the plot will be drawn.
        position : tuple
            Position (z, y) in which the patch will be drawn.
        length : float
            minimum length of shaft elements

        Returns
        -------
        """
        zpos, ypos = position
        le = length

        # bokeh plot - upper bearing visual representation
        bk_ax.quad(
            top=-ypos + le / 3,
            bottom=-ypos,
            left=zpos - le / 6,
            right=zpos + le / 6,
            line_color=bokeh_colors[0],
            line_width=1,
            fill_alpha=1,
            fill_color=bokeh_colors[1],
            legend="Bearing",
        )
        # bokeh plot - lower bearing visual representation
        bk_ax.quad(
            top=ypos,
            bottom=ypos - le / 3,
            left=zpos - le / 6,
            right=zpos + le / 6,
            line_color=bokeh_colors[0],
            line_width=1,
            fill_alpha=1,
            fill_color=bokeh_colors[1],
        )

    @classmethod
    def table_to_toml(cls, n, file):
        """Convert a table with parameters of a bearing element to a dictionary ready to save
        to a toml file that can be later loaded by ross.
        Parameters
        ----------
        n : int
            The node in which the bearing will be located in the rotor.
        file: str
            Path to the file containing the bearing parameters.
        Returns
        -------
        dict
            A dict that is ready to save to toml and readable by ross.
        """
        b_elem = cls.from_table(n, file)
        data = {
            "n": b_elem.n,
            "kxx": b_elem.kxx.coefficient,
            "cxx": b_elem.cxx.coefficient,
            "kyy": b_elem.kyy.coefficient,
            "kxy": b_elem.kxy.coefficient,
            "kyx": b_elem.kyx.coefficient,
            "cyy": b_elem.cyy.coefficient,
            "cxy": b_elem.cxy.coefficient,
            "cyx": b_elem.cyx.coefficient,
            "w": b_elem.w,
        }
        return data

    @classmethod
    def from_table(cls, n, file):
        """Instantiate a bearing using inputs from a table, either excel or csv.
        Parameters
        ----------
        n : int
            The node in which the bearing will be located in the rotor.
        file: str
            Path to the file containing the bearing parameters.
        Returns
        -------
        A bearing object.
        """
        try:
            df = pd.read_excel(file, header=None)
        except FileNotFoundError:
            sys.exit(file + " not found.")
        except xlrd.biffh.XLRDError:
            df = pd.read_csv(file, header=None)
        header_index = -1
        header_found = False
        for index, row in df.iterrows():
            for i in range(0, row.size):
                if isinstance(row[i], str):
                    if row[i].lower() == 'kxx':
                        header_index = index
                        header_found = True
                        break
            if header_found:
                break
        if header_index < 0:
            sys.exit("Could not find the header. Make sure the sheet has a header "
                     "containing the names of the columns.")
        try:
            df = pd.read_excel(file, header=header_index)
        except xlrd.biffh.XLRDError:
            df = pd.read_csv(file, header=header_index)
        first_data_row = -1
        for index, row in df.iterrows():
            if isinstance(row[0], int) or isinstance(row[0], float):
                first_data_row = index
                break
        if first_data_row < 0:
            sys.exit("Could not find the data. Make sure you have at least one row containing "
                     "data below the header.")
        for i in range(0, first_data_row):
            df = df.drop(i)
        nan_found = False
        for index, row in df.iterrows():
            for i in range(first_data_row, row.size):
                if pd.isna(row[i]):
                    nan_found = True
                    row[i] = 0
        if nan_found:
            warnings.warn("One or more NaN found. They were replaced with zeros.")
        parameters = []
        possible_names = [["kxx", "Kxx", "KXX"], ["cxx", "Cxx", "CXX"], ["kyy", "Kyy", "KYY"],
                          ["kxy", "Kxy", "KXY"], ["kyx", "Kyx", "KYX"], ["cyy", "Cyy", "CYY"],
                          ["cxy", "Cxy", "CXY"], ["cyx", "Cyx", "CYX"],
                          ["w", "W", "speed", "Speed", "SPEED"]]
        for name_group in possible_names:
            for name in name_group:
                try:
                    parameter = df[name].tolist()
                    parameters.append(parameter)
                    break
                except KeyError:
                    continue
        return cls(
            n,
            kxx=parameters[0],
            cxx=parameters[1],
            kyy=parameters[2],
            kxy=parameters[3],
            kyx=parameters[4],
            cyy=parameters[5],
            cxy=parameters[6],
            cyx=parameters[7],
            w=parameters[8],
        )

    @classmethod
    def from_fluid_flow(
        cls,
        n,
        nz,
        ntheta,
        nradius,
        length,
        omega,
        p_in,
        p_out,
        radius_rotor,
        radius_stator,
        visc,
        rho,
        eccentricity=None,
        load=None,
    ):
        """Instantiate a bearing using inputs from its fluid flow.
        Parameters
        ----------
        n : int
            The node in which the bearing will be located in the rotor.
        Grid related
        ^^^^^^^^^^^^
        Describes the discretization of the problem
        nz: int
            Number of points along the Z direction (direction of flow).
        ntheta: int
            Number of points along the direction theta. NOTE: ntheta must be odd.
        nradius: int
            Number of points along the direction r.
        length: float
            Length in the Z direction (m).

        Operation conditions
        ^^^^^^^^^^^^^^^^^^^^
        Describes the operation conditions.
        omega: float
            Rotation of the rotor (rad/s).
        p_in: float
            Input Pressure (Pa).
        p_out: float
            Output Pressure (Pa).
        load: float
            Load applied to the rotor (N).

        Geometric data of the problem
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
        Describes the geometric data of the problem.
        radius_rotor: float
            Rotor radius (m).
        radius_stator: float
            Stator Radius (m).
        eccentricity: float
            Eccentricity (m) is the euclidean distance between rotor and stator centers.
            The center of the stator is in position (0,0).

        Fluid characteristics
        ^^^^^^^^^^^^^^^^^^^^^
        Describes the fluid characteristics.
        visc: float
            Viscosity (Pa.s).
        rho: float
            Fluid density(Kg/m^3).
        Returns
        -------
        A bearing object.
        """
        fluid_flow = flow.PressureMatrix(
            nz,
            ntheta,
            nradius,
            length,
            omega,
            p_in,
            p_out,
            radius_rotor,
            radius_stator,
            visc,
            rho,
            eccentricity=eccentricity,
            load=load,
        )
        k = fluid_flow.get_analytical_damping_matrix()
        c = fluid_flow.get_analytical_stiffness_matrix()
        return cls(
            n,
            kxx=k[0],
            cxx=c[0],
            kyy=k[3],
            kxy=k[1],
            kyx=k[2],
            cyy=c[3],
            cxy=c[1],
            cyx=c[2],
            w=fluid_flow.omega,
        )


class SealElement(BearingElement):
    def __init__(
        self,
        n,
        kxx,
        cxx,
        kyy=None,
        kxy=0,
        kyx=0,
        cyy=None,
        cxy=0,
        cyx=0,
        w=None,
        seal_leakage=None,
    ):
        super().__init__(
            n=n,
            w=w,
            kxx=kxx,
            kxy=kxy,
            kyx=kyx,
            kyy=kyy,
            cxx=cxx,
            cxy=cxy,
            cyx=cyx,
            cyy=cyy,
        )

        self.seal_leakage = seal_leakage
        self.color = "#77ACA2"

    def patch(self, position, ax):
        """Seal element patch.
        Patch that will be used to draw the seal element.
        Parameters
        ----------
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        bk_ax : bokeh plotting axes, optional
            Axes in which the plot will be drawn.
        position : tuple
            Position in which the patch will be drawn.
        Returns
        -------
        ax : matplotlib axes
            Returns the axes object with the plot.
        bk_ax : bokeh plotting axes
            Returns the axes object with the plot.
        """
        zpos, ypos = position
        hw = 0.05

        #  node (x pos), outer diam. (y pos)
        seal_points_u = [
            [zpos, ypos * 1.1],  # upper
            [zpos + hw, ypos * 1.1],
            [zpos + hw, ypos * 1.3],
            [zpos - hw, ypos * 1.3],
            [zpos - hw, ypos * 1.1],
            [zpos, ypos * 1.1],
        ]
        seal_points_l = [
            [zpos, -ypos * 1.1],  # lower
            [zpos + hw, -(ypos * 1.1)],
            [zpos + hw, -(ypos * 1.3)],
            [zpos - hw, -(ypos * 1.3)],
            [zpos - hw, -(ypos * 1.1)],
            [zpos, -ypos * 1.1],
        ]
        ax.add_patch(mpatches.Polygon(seal_points_u, facecolor=self.color))
        ax.add_patch(mpatches.Polygon(seal_points_l, facecolor=self.color))

    def bokeh_patch(self, position, bk_ax):
        """Seal element patch.
        Patch that will be used to draw the seal element.
        Parameters
        ----------
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        bk_ax : bokeh plotting axes, optional
            Axes in which the plot will be drawn.
        position : tuple
            Position in which the patch will be drawn.
        Returns
        -------
        ax : matplotlib axes
            Returns the axes object with the plot.
        bk_ax : bokeh plotting axes
            Returns the axes object with the plot.
        """
        zpos, ypos = position
        hw = 0.05

        # bokeh plot - node (x pos), outer diam. (y pos)
        bk_seal_points_u = [
            [zpos, zpos + hw, zpos + hw, zpos - hw, zpos - hw],
            [ypos * 1.1, ypos * 1.1, ypos * 1.3, ypos * 1.3, ypos * 1.1],
        ]

        bk_seal_points_l = [
            [zpos, zpos + hw, zpos + hw, zpos - hw, zpos - hw],
            [ypos * 1.1, ypos * 1.1, ypos * 1.3, ypos * 1.3, ypos * 1.1],
        ]

        # bokeh plot - plot disks elements
        bk_ax.patch(
            bk_seal_points_u[0],
            bk_seal_points_u[1],
            alpha=0.5,
            line_width=2,
            color=bokeh_colors[6],
        )

        bk_ax.patch(
            bk_seal_points_l[0],
            bk_seal_points_l[1],
            alpha=0.5,
            line_width=2,
            color=bokeh_colors[6],
        )



import sys
import warnings
import matplotlib.pyplot as plt
import numpy as np
from bokeh.plotting import figure, output_file


class PressureMatrix:
    r"""This class calculates the pressure matrix of the object with the given parameters.

    It is supposed to be an attribute of a bearing element,
    but can work on its on to provide graphics for the user.

    Parameters
    ----------
    Grid related
    ^^^^^^^^^^^^
    Describes the discretization of the problem
    nz: int
        Number of points along the Z direction (direction of flow).
    ntheta: int
        Number of points along the direction theta. NOTE: ntheta must be odd.
    nradius: int
        Number of points along the direction r.
    length: float
        Length in the Z direction (m).

    Operation conditions
    ^^^^^^^^^^^^^^^^^^^^
    Describes the operation conditions.
    omega: float
        Rotation of the rotor (rad/s).
    p_in: float
        Input Pressure (Pa).
    p_out: float
        Output Pressure (Pa).
    load: float
        Load applied to the rotor (N).

    Geometric data of the problem
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    Describes the geometric data of the problem.
    radius_rotor: float
        Rotor radius (m).
    radius_stator: float
        Stator Radius (m).
    eccentricity: float, optional
        Eccentricity (m) is the euclidean distance between rotor and stator centers.
        The center of the stator is in position (0,0).
    beta: float, optional
        Attitude angle. Angle between the origin and the eccentricity (rad).

    Fluid characteristics
    ^^^^^^^^^^^^^^^^^^^^^
    Describes the fluid characteristics.
    viscosity: float
        Viscosity (Pa.s).
    density: float
        Fluid density(Kg/m^3).

    Returns
    -------
    An object containing the pressure matrix and its data.

    Attributes
    ----------
    ltheta: float
        Length in the theta direction (rad).
    dz: float
        Range size in the Z direction.
    dtheta: float
        Range size in the theta direction.
    ntotal: int
        Number of nodes in the grid., ntheta, nradius, n_interv_z, n_interv_theta,
    n_interv_z: int
        Number of intervals on Z.
    n_interv_theta: int
      Number of intervals on theta.
    n_interv_radius: int
        Number of intervals on r.
    p_mat_analytical : array of shape (nz, ntheta)
        The analytical pressure matrix.
    p_mat_numerical : array of shape (nz, ntheta)
        The numerical pressure matrix.
    xi: float
        Eccentricity (m) (distance between rotor and stator centers) on the x-axis.
        It is the position of the center of the rotor.
        The center of the stator is in position (0,0).
    yi: float
        Eccentricity (m) (distance between rotor and stator centers) on the y-axis.
        It is the position of the center of the rotor.
        The center of the stator is in position (0,0).
    re : array of shape (nz, ntheta)
        The external radius in each position of the grid.
    ri : array of shape (nz, ntheta)
        The internal radius in each position of the grid.
    z : array of shape (1, nz)
        z along the object. It goes from 0 to lb.
    xre : array of shape (nz, ntheta)
        x value of the external radius.
    xri : array of shape (nz, ntheta)
        x value of the internal radius.
    yre : array of shape (nz, ntheta)
        y value of the external radius.
    yri : array of shape (nz, ntheta)
        y value of the internal radius.
    eccentricity : float
        distance between the center of the rotor and the stator.
    difference_between_radius: float
        distance between the radius of the stator and the radius of the rotor.
    eccentricity_ratio: float
        eccentricity/difference_between_radius
    bearing_type: str
        type of structure. 'short_bearing': short; 'long_bearing': long;
        'medium_size': in between short and long.
        if length/radius_stator <= 1/4 it is short.
        if length/radius_stator > 8 it is long.
    radial_clearance: float
        Difference between both stator and rotor radius, regardless of eccentricity.
    analytical_pressure_matrix_available: bool
        True if analytically calculated pressure matrix is available.
    numerical_pressure_matrix_available: bool
        True if numerically calculated pressure matrix is available.

    Examples
    --------
    >>> from ross.fluid_flow import fluid_flow as flow
    >>> import numpy as np
    >>> from bokeh.plotting import show
    >>> nz = 8
    >>> ntheta = 64
    >>> nradius = 11
    >>> length = 0.01
    >>> omega = 100.*2*np.pi/60
    >>> p_in = 0.
    >>> p_out = 0.
    >>> radius_rotor = 0.08
    >>> radius_stator = 0.1
    >>> viscosity = 0.015
    >>> density = 860.
    >>> eccentricity = 0.001
    >>> beta = np.pi
    >>> my_fluid_flow = flow.PressureMatrix(nz, ntheta, nradius, length,
    ...                                          omega, p_in, p_out, radius_rotor,
    ...                                          radius_stator, viscosity, density, beta=beta, eccentricity=eccentricity)
    >>> my_fluid_flow.calculate_pressure_matrix_analytical() # doctest: +ELLIPSIS
    array([[-0.00000...
    >>> my_fluid_flow.calculate_pressure_matrix_numerical() # doctest: +ELLIPSIS
    array([[...
    >>> # to show the plots you can use:
    >>> # show(my_fluid_flow.plot_eccentricity())
    >>> # show(my_fluid_flow.plot_pressure_theta(z=int(nz/2)))
    >>> my_fluid_flow.matplot_pressure_theta(z=int(nz/2)) # doctest: +ELLIPSIS
    <matplotlib.axes._subplots.AxesSubplot object at...
    """

    def __init__(
        self,
        nz,
        ntheta,
        nradius,
        length,
        omega,
        p_in,
        p_out,
        radius_rotor,
        radius_stator,
        viscosity,
        density,
        beta=2.356,
        eccentricity=None,
        load=None,
    ):
        if load is None and eccentricity is None:
            sys.exit("Either load or eccentricity must be given.")
        self.nz = nz
        self.ntheta = ntheta
        self.nradius = nradius
        self.n_interv_z = nz - 1
        self.n_interv_theta = ntheta - 1
        self.n_interv_radius = nradius - 1
        self.length = length
        self.ltheta = 2.0 * np.pi
        self.dz = length / self.n_interv_z
        self.dtheta = self.ltheta / self.n_interv_theta
        self.ntotal = self.nz * self.ntheta
        self.omega = omega
        self.p_in = p_in
        self.p_out = p_out
        self.radius_rotor = radius_rotor
        self.radius_stator = radius_stator
        self.viscosity = viscosity
        self.density = density
        self.radial_clearance = self.radius_stator - self.radius_rotor
        self.difference_between_radius = radius_stator - radius_rotor
        self.bearing_type = ""
        if self.length / self.radius_stator <= 1 / 4:
            self.bearing_type = "short_bearing"
        elif self.length / self.radius_stator > 8:
            self.bearing_type = "long_bearing"
        else:
            self.bearing_type = "medium_size"
        self.eccentricity = eccentricity
        self.beta = beta
        self.eccentricity_ratio = None
        self.load = load
        if self.eccentricity is None:
            self.eccentricity = (
                self.calculate_eccentricity_ratio() * self.difference_between_radius
            )
        self.eccentricity_ratio = self.eccentricity / self.difference_between_radius
        if self.load is None:
            self.load = self.get_rotor_load()
        self.xi = self.eccentricity * np.cos(2*np.pi - beta)
        self.yi = self.eccentricity * np.sin(2*np.pi - beta)
        self.re = np.zeros([self.nz, self.ntheta])
        self.ri = np.zeros([self.nz, self.ntheta])
        self.z = np.zeros([1, self.nz])
        self.xre = np.zeros([self.nz, self.ntheta])
        self.xri = np.zeros([self.nz, self.ntheta])
        self.yre = np.zeros([self.nz, self.ntheta])
        self.yri = np.zeros([self.nz, self.ntheta])
        self.p_mat_analytical = np.zeros([self.nz, self.ntheta])
        self.c1 = np.zeros([self.nz, self.ntheta])
        self.c2 = np.zeros([self.nz, self.ntheta])
        self.c0w = np.zeros([self.nz, self.ntheta])
        self.M = np.zeros([self.ntotal, self.ntotal])
        self.f = np.zeros([self.ntotal, 1])
        self.P = np.zeros([self.ntotal, 1])
        self.p_mat_numerical = np.zeros([self.nz, self.ntheta])
        self.calculate_coefficients()
        self.analytical_pressure_matrix_available = False
        self.numerical_pressure_matrix_available = False

    def calculate_pressure_matrix_analytical(self):
        """This function calculates the pressure matrix analytically, based on the book Tribology Series vol. 33, by
        Frene et al., chapter 5.
        """
        if self.bearing_type == "short_bearing":
            for i in range(self.nz):
                for j in range(self.ntheta):
                    # fmt: off
                    self.p_mat_analytical[i][j] = (((-3*self.viscosity*self.omega)/self.difference_between_radius**2) *
                                                   ((i * self.dz - (self.length / 2)) ** 2 - (self.length ** 2) / 4) *
                                                   (self.eccentricity_ratio * np.sin(j * self.dtheta)) /
                                                   (1 + self.eccentricity_ratio * np.cos(j * self.dtheta))**3)
                    # fmt: on
        self.analytical_pressure_matrix_available = True
        return self.p_mat_analytical

    def calculate_pressure_matrix_analytical2(self):
        """This function calculates the pressure matrix analytically, based on the chapter Linear and Nonlinear
        Rotordynamics, by Ishida and Yamamoto, from the book Flow-Induced Vibrations.
        """
        if self.bearing_type == "short_bearing":
            for i in range(self.nz):
                for j in range(self.ntheta):
                    # fmt: off
                    self.p_mat_analytical[i][j] = (3 * self.viscosity / ((self.difference_between_radius ** 2) *
                                                                    (1. + self.eccentricity_ratio * np.cos(
                                                                        j * self.dtheta)) ** 3)) * \
                                                  (-self.eccentricity_ratio * self.omega * np.sin(j * self.dtheta)) * \
                                                  (((i * self.dz - (self.length / 2)) ** 2) - (self.length ** 2) / 4)
                    # fmt: on
        self.analytical_pressure_matrix_available = True

    def calculate_coefficients(self):
        """This function calculates the constants that form the Poisson equation
        of the discrete pressure (central differences in the second
        derivatives). It is executed when the class is instantiated.
        """

        for i in range(self.nz):
            zno = i * self.dz
            self.z[0][i] = zno
            plot_eccentricity_error = False
            position = -1
            for j in range(self.ntheta):
                # fmt: off
                gama = j * self.dtheta + (np.pi - self.beta)
                [radius_external, self.xre[i][j], self.yre[i][j]] =\
                    self.external_radius_function(gama)
                [radius_internal, self.xri[i][j], self.yri[i][j]] =\
                    self.internal_radius_function(zno, gama)
                self.re[i][j] = radius_external
                self.ri[i][j] = radius_internal

                w = self.omega * self.ri[i][j]

                k = (self.re[i][j] ** 2 * (np.log(self.re[i][j]) - 1 / 2) - self.ri[i][j] ** 2 *
                     (np.log(self.ri[i][j]) - 1 / 2)) / (self.ri[i][j] ** 2 - self.re[i][j] ** 2)

                self.c1[i][j] = (1 / (4 * self.viscosity)) * ((self.re[i][j] ** 2 * np.log(self.re[i][j]) -
                                                          self.ri[i][j] ** 2 * np.log(self.ri[i][j]) +
                                                          (self.re[i][j] ** 2 - self.ri[i][j] ** 2) *
                                                          (k - 1)) - 2 * self.re[i][j] ** 2 *
                                                         ((np.log(self.re[i][j]) + k - 1 / 2) * np.log(self.re[i][j] /
                                                                                                       self.ri[i][j])))

                self.c2[i][j] = (- self.ri[i][j] ** 2) / (8 * self.viscosity) *\
                                ((self.re[i][j] ** 2 - self.ri[i][j] ** 2 -
                                  (self.re[i][j] ** 4 - self.ri[i][j] ** 4) /
                                  (2 * self.ri[i][j] ** 2)) +
                                 ((self.re[i][j] ** 2 - self.ri[i][j] ** 2) /
                                  (self.ri[i][j] ** 2 *
                                   np.log(self.re[i][j] / self.ri[i][j]))) *
                                 (self.re[i][j] ** 2 * np.log(self.re[i][j] / self.ri[i][j]) -
                                  (self.re[i][j] ** 2 - self.ri[i][j] ** 2) / 2))

                self.c0w[i][j] = (- w * self.ri[i][j] *
                                  (np.log(self.re[i][j] / self.ri[i][j]) *
                                  (1 + (self.ri[i][j] ** 2) / (self.re[i][j] ** 2 - self.ri[i][j] ** 2)) - 1 / 2))
                # fmt: on
                if not plot_eccentricity_error:
                    if abs(self.xri[i][j]) > abs(self.xre[i][j]) or abs(
                        self.yri[i][j]
                    ) > abs(self.yre[i][j]):
                        plot_eccentricity_error = True
                        position = i
            if plot_eccentricity_error:
                self.plot_eccentricity(position)
                sys.exit(
                    "Error: The given parameters create a rotor that is not inside the stator. "
                    "Check the plotted figure and fix accordingly."
                )

    def mounting_matrix(self):
        """This function assembles the matrix M and the independent vector f
        """

        # fmt: off
        count = 0
        for x in range(self.ntheta):
            self.M[count][count] = 1
            self.f[count][0] = self.p_in
            count = count + self.nz - 1
            self.M[count][count] = 1
            self.f[count][0] = self.p_out
            count = count + 1
        count = 0
        for x in range(self.nz - 2):
            self.M[self.ntotal - self.nz + 1 + count][1 + count] = 1
            self.M[self.ntotal - self.nz + 1 + count][self.ntotal - self.nz + 1 + count] = -1
            count = count + 1
        count = 1
        j = 0
        for i in range(1, self.nz - 1):
            a = (1/self.dtheta**2)*(self.c1[i][self.ntheta-1])
            self.M[count][self.ntotal - 2 * self.nz + count] = a
            b = (1 / self.dz ** 2) * (self.c2[i - 1, j])
            self.M[count][count - 1] = b
            c = -((1/self.dtheta**2) * ((self.c1[i][j]) + self.c1[i][self.ntheta-1])
                  + (1/self.dz**2) * (self.c2[i][j] + self.c2[i-1][j]))
            self.M[count, count] = c
            d = (1 / self.dz ** 2) * (self.c2[i][j])
            self.M[count][count + 1] = d
            e = (1/self.dtheta**2) * (self.c1[i][j])
            self.M[count][count + self.nz] = e
            count = count + 1
        count = self.nz + 1
        for j in range(1, self.ntheta - 1):
            for i in range(1, self.nz - 1):
                a = (1/self.dtheta**2) * (self.c1[i, j-1])
                self.M[count][count - self.nz] = a
                b = (1 / self.dz ** 2) * (self.c2[i - 1][j])
                self.M[count][count - 1] = b
                c = -((1 / self.dtheta**2) * ((self.c1[i][j]) + self.c1[i][j-1])
                      + (1 / self.dz**2) * (self.c2[i][j] + self.c2[i-1][j]))
                self.M[count, count] = c
                d = (1 / self.dz ** 2) * (self.c2[i][j])
                self.M[count][count + 1] = d
                e = (1 / self.dtheta**2) * (self.c1[i][j])
                self.M[count][count + self.nz] = e
                count = count + 1
            count = count + 2
        count = 1
        for j in range(self.ntheta - 1):
            for i in range(1, self.nz - 1):
                if j == 0:
                    self.f[count][0] = (self.c0w[i][j] - self.c0w[i][self.ntheta - 1]) / self.dtheta
                else:
                    self.f[count][0] = (self.c0w[i, j] - self.c0w[i, j - 1]) / self.dtheta
                count = count + 1
            count = count + 2
        # fmt: on

    def resolves_matrix(self):
        """This function resolves the linear system [M]{P}={f}.
        """
        self.P = np.linalg.solve(self.M, self.f)

    def calculate_pressure_matrix_numerical(self):
        """This function calculates the pressure matrix numerically.
        """
        self.mounting_matrix()
        self.resolves_matrix()
        for i in range(self.nz):
            for j in range(self.ntheta):
                k = j * self.nz + i
                self.p_mat_numerical[i][j] = self.P[k]
        self.numerical_pressure_matrix_available = True
        return self.p_mat_numerical

    def internal_radius_function(self, z, gama):
        """This function calculates the radius of the rotor given the
        radius and the position z.
        Parameters
        ----------
        z: float
            Distance along the z-axis.
        gama: float
            Gama is the distance in the theta-axis. It should range from 0 to 2*np.pi.
        Returns
        -------
        radius_internal: float
            The size of the internal radius at that point.
        xri: float
            The position x of the returned internal radius.
        yri: float
            The position y of the returned internal radius.
        """
        if gama > (np.pi - self.beta) and gama < (2*np.pi - self.beta):
            alpha = np.absolute(2*np.pi - gama - self.beta)
            radius_internal = np.sqrt(self.radius_rotor ** 2
                                      - (self.eccentricity * np.sin(alpha)) ** 2) + self.eccentricity * np.cos(alpha)
            xri = radius_internal * np.cos(gama)
            yri = radius_internal * np.sin(gama)

        else:
            alpha = self.beta + gama
            radius_internal = np.sqrt(self.radius_rotor ** 2
                                      - (self.eccentricity * np.sin(alpha)) ** 2) + self.eccentricity * np.cos(alpha)
            xri = radius_internal * np.cos(gama)
            yri = radius_internal * np.sin(gama)

        return radius_internal, xri, yri

    def external_radius_function(self, gama):
        """This function calculates the radius of the stator.
        Parameters
        ----------
        gama: float
            Gama is the distance in the theta-axis. It should range from 0 to 2*np.pi.
        Returns
        -------
        radius_external: float
            The size of the external radius at that point.
        xre: float
            The position x of the returned external radius.
        yre: float
            The position y of the returned external radius.
        """
        radius_external = self.radius_stator
        xre = radius_external * np.cos(gama)
        yre = radius_external * np.sin(gama)

        return radius_external, xre, yre

    def get_rotor_load(self):
        """Returns the load applied to the rotor, based on the eccentricity ratio.
        Suitable only for short bearings.
        Returns
        -------
        float
            Load applied to the rotor.
        """
        if not self.bearing_type == "short_bearing":
            warnings.warn(
                "Function get_rotor_load suitable only for short bearings. "
                "The ratio between the bearing length and its radius should be less or "
                "equal to 0.25. Currently we have "
                + str(self.length / self.radius_stator)
                + "."
            )
        return (
            (
                np.pi
                * self.radius_stator
                * 2
                * self.omega
                * self.viscosity
                * (self.length ** 3)
                * self.eccentricity_ratio
            )
            / (
                8
                * (self.radial_clearance ** 2)
                * ((1 - self.eccentricity_ratio ** 2) ** 2)
            )
        ) * (np.sqrt((16 / (np.pi ** 2) - 1) * self.eccentricity_ratio ** 2 + 1))

    def modified_sommerfeld_number(self):
        """Returns the modified sommerfeld number.
        Returns
        -------
        float
            The modified sommerfeld number.
        """
        return (
            self.radius_stator * 2 * self.omega * self.viscosity * (self.length ** 3)
        ) / (8 * self.load * (self.radial_clearance ** 2))

    def sommerfeld_number(self):
        """Returns the sommerfeld number.
        Returns
        -------
        float
            The sommerfeld number.
        """
        modified_s = self.modified_sommerfeld_number()
        return (modified_s / np.pi) * (self.radius_stator * 2 / self.length) ** 2

    def calculate_eccentricity_ratio(self):
        """Calculate the eccentricity ratio for a given load using the sommerfeld number.
        Suitable only for short bearings.
        Returns
        -------
        float
            The eccentricity ratio.
        """
        if not self.bearing_type == "short_bearing":
            warnings.warn(
                "Function calculate_eccentricity_ratio suitable only for short bearings. "
                "The ratio between the bearing length and its radius should be less or "
                "equal to 0.25. Currently we have "
                + str(self.length / self.radius_stator)
                + "."
            )
        s = self.modified_sommerfeld_number()
        coefficients = [
            1,
            -4,
            (6 - (s ** 2) * (16 - np.pi ** 2)),
            -(4 + (np.pi ** 2) * (s ** 2)),
            1,
        ]
        roots = np.roots(coefficients)
        for i in range(0, len(roots)):
            if 0 <= roots[i] <= 1:
                return np.sqrt(roots[i].real)
        sys.exit("Eccentricity ratio could not be calculated.")

    def get_analytical_stiffness_matrix(self):
        """Returns the stiffness matrix calculated analytically.
        Suitable only for short bearings.
        Returns
        -------
        list of floats
            A list of length four including stiffness floats in this order: kxx, kxy, kyx, kyy
        """
        if not self.bearing_type == "short_bearing":
            warnings.warn(
                "Function get_analytical_stiffness_matrix suitable only for short bearings. "
                "The ratio between the bearing length and its radius should be less or "
                "equal to 0.25. Currently we have "
                + str(self.length / self.radius_stator)
                + "."
            )
        # fmt: off
        f = self.get_rotor_load()
        h0 = 1.0/(((np.pi**2)*(1 - self.eccentricity_ratio**2) + 16*self.eccentricity_ratio**2)**1.5)
        a = f/self.radial_clearance
        kxx = a*h0*4*((np.pi**2)*(2 - self.eccentricity_ratio**2) + 16*self.eccentricity_ratio**2)
        kxy = (a*h0*np.pi*((np.pi**2)*(1 - self.eccentricity_ratio**2)**2 -
               16*self.eccentricity_ratio**4) /
               (self.eccentricity_ratio*np.sqrt(1 - self.eccentricity_ratio**2)))
        kyx = (-a*h0*np.pi*((np.pi**2)*(1 - self.eccentricity_ratio**2) *
               (1 + 2*self.eccentricity_ratio**2) +
               (32*self.eccentricity_ratio**2)*(1 + self.eccentricity_ratio**2)) /
               (self.eccentricity_ratio*np.sqrt(1 - self.eccentricity_ratio**2)))
        kyy = (a*h0*4*((np.pi**2)*(1 + 2*self.eccentricity_ratio**2) +
                                  ((32*self.eccentricity_ratio**2) *
                                   (1 + self.eccentricity_ratio**2))/(1 - self.eccentricity_ratio**2)))
        # fmt: on
        return [kxx, kxy, kyx, kyy]

    def get_analytical_damping_matrix(self):
        """Returns the damping matrix calculated analytically.
        Suitable only for short bearings.
        Returns
        -------
        list of floats
            A list of length four including stiffness floats in this order: cxx, cxy, cyx, cyy
        """
        if not self.bearing_type == "short_bearing":
            warnings.warn(
                "Function get_analytical_damping_matrix suitable only for short bearings. "
                "The ratio between the bearing length and its radius should be less or "
                "equal to 0.25. Currently we have "
                + str(self.length / self.radius_stator)
                + "."
            )
        # fmt: off
        f = self.get_rotor_load()
        h0 = 1.0/(((np.pi**2)*(1 - self.eccentricity_ratio**2) + 16*self.eccentricity_ratio**2)**1.5)
        a = f/(self.radial_clearance * self.omega)
        cxx = (a*h0*2*np.pi*np.sqrt(1 - self.eccentricity_ratio**2) *
               ((np.pi**2)*(1 + 2*self.eccentricity_ratio**2) - 16*self.eccentricity_ratio**2)/self.eccentricity_ratio)
        cxy = (-a*h0*8*((np.pi**2)*(1 + 2*self.eccentricity_ratio**2) - 16*self.eccentricity_ratio**2))
        cyx = cxy
        cyy = (a*h0*(2*np.pi*((np.pi**2)*(1 - self.eccentricity_ratio**2)**2 + 48*self.eccentricity_ratio**2)) /
               (self.eccentricity_ratio*np.sqrt(1 - self.eccentricity_ratio**2)))
        # fmt: on
        return [cxx, cxy, cyx, cyy]

    def plot_eccentricity(self, z=0):
        """This function assembles pressure graphic along the z-axis.
        The first few plots are of a different color to indicate where theta begins.
        Parameters
        ----------
        z: int, optional
            The distance in z where to cut and plot.
        Returns
        -------
        Figure
            An object containing the plot.
        """
        p = figure(
            title="Cut in plane Z=" + str(z),
            x_axis_label="X axis",
            y_axis_label="Y axis",
        )
        for j in range(0, self.ntheta):
            p.circle(self.xre[z][j], self.yre[z][j], color="red")
            p.circle(self.xri[z][j], self.yri[z][j], color="blue")
            p.circle(0, 0, color="blue")
            p.circle(self.xi, self.yi, color="red")
        p.circle(0, 0, color="black")
        return p

    def plot_pressure_z(self, theta=0):
        """This function assembles pressure graphic along the z-axis for one or both the
        numerically (blue) and analytically (red) calculated pressure matrices, depending on if
        one or both were calculated.
        Parameters
        ----------
        theta: int, optional
            The theta to be considered.
        Returns
        -------
        Figure
            An object containing the plot.
        """
        if (
            not self.numerical_pressure_matrix_available
            and not self.analytical_pressure_matrix_available
        ):
            sys.exit(
                "Must calculate the pressure matrix. "
                "Try calling calculate_pressure_matrix_numerical() or calculate_pressure_matrix_analytical() first."
            )
        x = []
        y_n = []
        y_a = []
        for i in range(0, self.nz):
            x.append(i * self.dz)
            y_n.append(self.p_mat_numerical[i][theta])
            y_a.append(self.p_mat_analytical[i][theta])
        p = figure(
            title="Pressure along the Z direction (direction of flow); Theta="
            + str(theta),
            x_axis_label="Points along Z",
        )
        if self.numerical_pressure_matrix_available:
            p.line(x, y_n, legend="Numerical pressure", color="blue", line_width=2)
        if self.analytical_pressure_matrix_available:
            p.line(x, y_a, legend="Analytical pressure", color="red", line_width=2)
        return p

    def plot_shape(self, theta=0):
        """This function assembles a graphic representing the geometry of the rotor.
        Parameters
        ----------
        theta: int, optional
            The theta to be considered.
        Returns
        -------
        Figure
            An object containing the plot.
        """
        x = np.zeros(self.nz)
        y_re = np.zeros(self.nz)
        y_ri = np.zeros(self.nz)
        for i in range(0, self.nz):
            x[i] = i * self.dz
            y_re[i] = self.re[i][theta]
            y_ri[i] = self.ri[i][theta]
        p = figure(
            title="Shapes of stator and rotor along Z; Theta=" + str(theta),
            x_axis_label="Points along Z",
            y_axis_label="Radial direction",
        )
        p.line(x, y_re, line_width=2, color="red")
        p.line(x, y_ri, line_width=2, color="blue")
        return p

    def plot_pressure_theta(self, z=0):
        """This function assembles pressure graphic in the theta direction for a given z
        for one or both the numerically (blue) and analytically (red) calculated pressure matrices,
        depending on if one or both were calculated.
        Parameters
        ----------
        z: int, optional
            The distance along z-axis to be considered.
        Returns
        -------
        Figure
            An object containing the plot.
        """
        if (
            not self.numerical_pressure_matrix_available
            and not self.analytical_pressure_matrix_available
        ):
            sys.exit(
                "Must calculate the pressure matrix. "
                "Try calling calculate_pressure_matrix_numerical() or calculate_pressure_matrix_analytical() first."
            )
        theta_list = []
        for theta in range(0, self.ntheta):
            theta_list.append(theta * self.dtheta)
        p = figure(
            title="Pressure along Theta; Z=" + str(z),
            x_axis_label="Points along Theta",
            y_axis_label="Pressure",
        )
        if self.numerical_pressure_matrix_available:
            p.line(
                theta_list,
                self.p_mat_numerical[z],
                legend="Numerical pressure",
                line_width=2,
                color="blue",
            )
        elif self.analytical_pressure_matrix_available:
            p.line(
                theta_list,
                self.p_mat_analytical[z],
                legend="Analytical pressure",
                line_width=2,
                color="red",
            )
        return p

    def matplot_eccentricity(self, z=0, ax=None):
        """This function assembles pressure graphic along the z-axis using matplotlib.
        The first few plots are of a different color to indicate where theta begins.
        Parameters
        ----------
        z: int, optional
            The distance in z where to cut and plot.
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        Returns
        -------
        ax : matplotlib axes
            Returns the axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()
        for j in range(0, self.ntheta):
            ax.plot(self.xre[z][j], self.yre[z][j], "r.")
            ax.plot(self.xri[z][j], self.yri[z][j], "b.")
            ax.plot(0, 0, "r*")
            ax.plot(self.xi, self.yi, "b*")
        ax.set_title("Cut in plane Z=" + str(z))
        ax.set_xlabel("X axis")
        ax.set_ylabel("Y axis")
        plt.axis("equal")
        return ax

    def matplot_pressure_z(self, theta=0, ax=None):
        """This function assembles pressure graphic along the z-axis using matplotlib
        for one or both the numerically (blue) and analytically (red) calculated pressure matrices,
        depending on if one or both were calculated.
        Parameters
        ----------
        theta: int, optional
            The distance in theta where to cut and plot.
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        Returns
        -------
        ax : matplotlib axes
            Returns the axes object with the plot.
        """
        if (
            not self.numerical_pressure_matrix_available
            and not self.analytical_pressure_matrix_available
        ):
            sys.exit(
                "Must calculate the pressure matrix. "
                "Try calling calculate_pressure_matrix_numerical() or calculate_pressure_matrix_analytical() first."
            )
        if ax is None:
            ax = plt.gca()
        x = np.zeros(self.nz)
        y_n = np.zeros(self.nz)
        y_a = np.zeros(self.nz)
        for i in range(0, self.nz):
            x[i] = i * self.dz
            y_n[i] = self.p_mat_numerical[i][theta]
            y_a[i] = self.p_mat_analytical[i][theta]
        if self.numerical_pressure_matrix_available:
            ax.plot(x, y_n, "b", label="Numerical pressure")
        if self.analytical_pressure_matrix_available:
            ax.plot(x, y_a, "r", label="Analytical pressure")
        ax.set_title(
            "Pressure along the Z direction (direction of flow); Theta=" + str(theta)
        )
        ax.set_xlabel("Points along Z")
        ax.set_ylabel("Pressure")
        return ax

    def matplot_shape(self, theta=0, ax=None):
        """This function assembles a graphic representing the geometry of the rotor using matplotlib.
        Parameters
        ----------
        theta: int, optional
            The theta to be considered.
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        Returns
        -------
        ax : matplotlib axes
            Returns the axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()
        x = np.zeros(self.nz)
        y_re = np.zeros(self.nz)
        y_ri = np.zeros(self.nz)
        for i in range(0, self.nz):
            x[i] = i * self.dz
            y_re[i] = self.re[i][theta]
            y_ri[i] = self.ri[i][theta]
        ax.plot(x, y_re, "r")
        ax.plot(x, y_ri, "b")
        ax.set_title("Shapes of stator and rotor along Z; Theta=" + str(theta))
        ax.set_xlabel("Points along Z")
        ax.set_ylabel("Radial direction")
        return ax

    def matplot_pressure_theta_cylindrical(self, z=0, from_numerical=True, ax=None):
        """This function assembles cylindrical pressure graphic in the theta direction for a given z,
        using matplotlib.
        Parameters
        ----------
        z: int, optional
            The distance along z-axis to be considered.
        from_numerical: bool, optional
            If True, takes the numerically calculated pressure matrix as entry.
            If False, takes the analytically calculated one instead.
            If condition cannot be satisfied (matrix not calculated), it will take the one that is available
            and raise a warning.
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        Returns
        -------
        ax : matplotlib axes
            Returns the axes object with the plot.
        """
        if (
            not self.numerical_pressure_matrix_available
            and not self.analytical_pressure_matrix_available
        ):
            sys.exit(
                "Must calculate the pressure matrix. "
                "Try calling calculate_pressure_matrix_numerical() or calculate_pressure_matrix_analytical() first."
            )
        p_mat = None
        if from_numerical:
            if self.numerical_pressure_matrix_available:
                p_mat = self.p_mat_numerical
            else:
                p_mat = self.p_mat_analytical
                warnings.warn(
                    "Plotting from analytically calculated pressure matrix, as numerically calculated "
                    "one is not available."
                )
        else:
            if self.analytical_pressure_matrix_available:
                p_mat = self.p_mat_analytical
            else:
                p_mat = self.p_mat_numerical
                warnings.warn(
                    "Plotting from numerically calculated pressure matrix, as analytically calculated "
                    "one is not available."
                )
        if ax is None:
            fig, ax = plt.subplots(subplot_kw=dict(projection="polar"))
        r = np.arange(
            0,
            self.radius_stator + 0.0001,
            (self.radius_stator - self.radius_rotor) / self.nradius,
        )
        theta = np.arange(-np.pi * 0.25, 1.75 * np.pi + self.dtheta / 2, self.dtheta)

        pressure_along_theta = np.zeros(self.ntheta)
        for i in range(0, self.ntheta):
            pressure_along_theta[i] = p_mat[0][i]

        min_pressure = np.amin(pressure_along_theta)

        r_matrix, theta_matrix = np.meshgrid(r, theta)
        z_matrix = np.zeros((theta.size, r.size))
        inner_radius_list = np.zeros(self.ntheta)
        pressure_list = np.zeros((theta.size, r.size))
        for i in range(0, theta.size):
            inner_radius = np.sqrt(
                self.xri[z][i] * self.xri[z][i] + self.yri[z][i] * self.yri[z][i]
            )
            inner_radius_list[i] = inner_radius
            for j in range(0, r.size):
                if r_matrix[i][j] < inner_radius:
                    continue
                pressure_list[i][j] = pressure_along_theta[i]
                z_matrix[i][j] = pressure_along_theta[i] - min_pressure + 0.01
        ax.contourf(theta_matrix, r_matrix, z_matrix, cmap="coolwarm")
        ax.set_title("Pressure along Theta; Z=" + str(z))
        return ax

    def matplot_pressure_theta(self, z=0, ax=None):
        """This function assembles pressure graphic in the theta direction for a given z,
        using matplotlib.
        Parameters
        ----------
        z: int, optional
            The distance along z-axis to be considered.
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        Returns
        -------
        ax : matplotlib axes
            Returns the axes object with the plot.
        """
        if (
            not self.numerical_pressure_matrix_available
            and not self.analytical_pressure_matrix_available
        ):
            sys.exit(
                "Must calculate the pressure matrix. "
                "Try calling calculate_pressure_matrix_numerical() or calculate_pressure_matrix_analytical() first."
            )
        if ax is None:
            ax = plt.gca()
        list_of_thetas = []
        for t in range(0, self.ntheta):
            list_of_thetas.append(t * self.dtheta)
        if self.numerical_pressure_matrix_available:
            ax.plot(
                list_of_thetas, self.p_mat_numerical[z], "b", label="Numerical pressure"
            )
        elif self.analytical_pressure_matrix_available:
            ax.plot(
                list_of_thetas,
                self.p_mat_analytical[z],
                "r",
                label="Analytical pressure",
            )
        ax.set_title("Pressure along Theta; Z=" + str(z))
        ax.set_xlabel("Points along Theta")
        ax.set_ylabel("Pressure")
        return ax



import pickle
import numpy as np
from scipy import interpolate

import matplotlib as mpl
import matplotlib.pyplot as plt
import bokeh.palettes as bp
from mpl_toolkits.mplot3d import Axes3D
from bokeh.layouts import gridplot
from bokeh.plotting import figure, show
from bokeh.transform import linear_cmap
from bokeh.models import (
        ColumnDataSource,
        ColorBar,
        Arrow,
        NormalHead,
        Label,
        HoverTool
)

# set bokeh palette of colors
bokeh_colors = bp.RdGy[11]


class Results(np.ndarray):
    """Class used to store results and provide plots.
    This class subclasses np.ndarray to provide additional info and a plot
    method to the calculated results from Rotor.
    Metadata about the results should be stored on info as a dictionary to be
    used on plot configurations and so on.
    Additional attributes can be passed as a dictionary in new_attributes kwarg.
    """

    def __new__(cls, input_array, new_attributes=None):
        obj = np.asarray(input_array).view(cls)

        for k, v in new_attributes.items():
            setattr(obj, k, v)

        # save new attributes names to create them on array finalize
        obj._new_attributes = new_attributes

        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return

        try:
            for k, v in obj._new_attributes.items():
                setattr(self, k, getattr(obj, k, v))
        except AttributeError:
            return

    def __reduce__(self):

        pickled_state = super().__reduce__()
        new_state = pickled_state[2] + (self._new_attributes,)

        return pickled_state[0], pickled_state[1], new_state

    def __setstate__(self, state):
        self._new_attributes = state[-1]
        for k, v in self._new_attributes.items():
            setattr(self, k, v)
        super().__setstate__(state[0:-1])

    def save(self, file):
        with open(file, mode="wb") as f:
            pickle.dump(self, f)

    def plot(self, *args, **kwargs):
        raise NotImplementedError


class CampbellResults:
    def __init__(self, speed_range, wd, log_dec, whirl_values):
        self.speed_range = speed_range
        self.wd = wd
        self.log_dec = log_dec
        self.whirl_values = whirl_values

    def _plot_matplotlib(self, harmonics=[1], fig=None, ax=None, **kwargs):
        if fig is None and ax is None:
            fig, ax = plt.subplots()

        wd = self.wd
        num_frequencies = wd.shape[1]
        log_dec = self.log_dec
        whirl = self.whirl_values
        speed_range = np.repeat(
            self.speed_range[:, np.newaxis], num_frequencies, axis=1
        )

        default_values = dict(cmap="RdBu", vmin=0.1, vmax=2.0, s=30, alpha=0.5)
        for k, v in default_values.items():
            kwargs.setdefault(k, v)

        for mark, whirl_dir, legend in zip(
            ["^", "o", "v"], [0.0, 0.5, 1.0], ["Foward", "Mixed", "Backward"]
        ):
            for i in range(num_frequencies):
                w_i = wd[:, i]
                whirl_i = whirl[:, i]
                log_dec_i = log_dec[:, i]
                speed_range_i = speed_range[:, i]

                for harm in harmonics:
                    ax.plot(
                        speed_range[:, 0],
                        harm * speed_range[:, 0],
                        color="k",
                        linewidth=1.5,
                        linestyle="-.",
                        alpha=0.75,
                        label="Rotor speed",
                    )

                    idx = np.argwhere(np.diff(np.sign(w_i - harm*speed_range_i))).flatten()
                    if len(idx) != 0:
                        idx = idx[0]

                        interpolated = interpolate.interp1d(
                            x=[speed_range_i[idx], speed_range_i[idx+1]],
                            y=[w_i[idx], w_i[idx+1]],
                            kind="linear"
                        )
                        xnew = np.linspace(
                            speed_range_i[idx],
                            speed_range_i[idx+1],
                            num=20,
                            endpoint=True,
                        )
                        ynew = interpolated(xnew)
                        idx = np.argwhere(np.diff(np.sign(ynew - harm*xnew))).flatten()

                        ax.scatter(xnew[idx], ynew[idx], marker="X", s=30, c="g")

                whirl_mask = whirl_i == whirl_dir
                if whirl_mask.shape[0] == 0:
                    continue
                else:
                    im = ax.scatter(
                        speed_range_i[whirl_mask],
                        w_i[whirl_mask],
                        c=log_dec_i[whirl_mask],
                        marker=mark,
                        **kwargs,
                    )

        if len(fig.axes) == 1:
            cbar = fig.colorbar(im)
            cbar.ax.set_ylabel("log dec")
            cbar.solids.set_edgecolor("face")

            forward_label = mpl.lines.Line2D(
                [], [], marker="^", lw=0, color="tab:blue", alpha=0.3, label="Forward"
            )
            backward_label = mpl.lines.Line2D(
                [], [], marker="v", lw=0, color="tab:blue", alpha=0.3, label="Backward"
            )
            mixed_label = mpl.lines.Line2D(
                [], [], marker="o", lw=0, color="tab:blue", alpha=0.3, label="Mixed"
            )

            legend = plt.legend(
                handles=[forward_label, backward_label, mixed_label], loc=2
            )

            ax.add_artist(legend)

            ax.set_xlabel("Rotor speed ($rad/s$)")
            ax.set_ylabel("Damped natural frequencies ($rad/s$)")

        return fig, ax

    def _plot_bokeh(self, harmonics=[1], **kwargs):
        wd = self.wd
        num_frequencies = wd.shape[1]
        log_dec = self.log_dec
        whirl = self.whirl_values
        speed_range = np.repeat(
            self.speed_range[:, np.newaxis], num_frequencies, axis=1
        )

        log_dec_map = log_dec.flatten()

        default_values = dict(
            cmap="viridis",
            vmin=min(log_dec_map),
            vmax=max(log_dec_map),
            s=30,
            alpha=1.0,
        )

        for k, v in default_values.items():
            kwargs.setdefault(k, v)

        camp = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            title="Campbell Diagram - Damped Natural Frequency Map",
            width=1600,
            height=900,
            x_axis_label="Rotor speed (rad/s)",
            y_axis_label="Damped natural frequencies (rad/s)",
        )
        camp.xaxis.axis_label_text_font_size = "14pt"
        camp.yaxis.axis_label_text_font_size = "14pt"

        color_mapper = linear_cmap(
                field_name="color",
                palette=bp.viridis(256),
                low=min(log_dec_map),
                high=max(log_dec_map),
        )

        for mark, whirl_dir, legend in zip(
            ["^", "o", "v"], [0.0, 0.5, 1.0], ["Foward", "Mixed", "Backward"]
        ):
            num_frequencies = wd.shape[1]
            for i in range(num_frequencies):
                w_i = wd[:, i]
                whirl_i = whirl[:, i]
                log_dec_i = log_dec[:, i]
                speed_range_i = speed_range[:, i]

                for harm in harmonics:
                    camp.line(
                        x=speed_range[:, 0],
                        y=harm * speed_range[:, 0],
                        line_width=3,
                        color=bokeh_colors[0],
                        line_dash="dotdash",
                        line_alpha=0.75,
                        legend="Rotor speed",
                        muted_color=bokeh_colors[0],
                        muted_alpha=0.2,
                    )

                    idx = np.argwhere(np.diff(np.sign(w_i - harm*speed_range_i))).flatten()
                    if len(idx) != 0:
                        idx = idx[0]

                        interpolated = interpolate.interp1d(
                            x=[speed_range_i[idx], speed_range_i[idx+1]],
                            y=[w_i[idx], w_i[idx+1]],
                            kind="linear"
                        )
                        xnew = np.linspace(
                            speed_range_i[idx],
                            speed_range_i[idx+1],
                            num=30,
                            endpoint=True,
                        )
                        ynew = interpolated(xnew)
                        idx = np.argwhere(np.diff(np.sign(ynew - harm*xnew))).flatten()

                        source = ColumnDataSource(
                                dict(xnew=xnew[idx], ynew=ynew[idx])
                        )
                        camp.square(
                                "xnew",
                                "ynew",
                                source=source,
                                size=10,
                                color=bokeh_colors[9],
                                name="critspeed"
                        )
                        hover = HoverTool(names=["critspeed"])
                        hover.tooltips = [
                                ("Frequency :", "@xnew"),
                                ("Critical Speed :", "@ynew")
                        ]
                        hover.mode = "mouse"

                whirl_mask = whirl_i == whirl_dir
                if whirl_mask.shape[0] == 0:
                    continue
                else:
                    source = ColumnDataSource(
                        dict(
                            x=speed_range_i[whirl_mask],
                            y=w_i[whirl_mask],
                            color=log_dec_i[whirl_mask],
                        )
                    )
                    camp.scatter(
                        x="x",
                        y="y",
                        color=color_mapper,
                        marker=mark,
                        fill_alpha=1.0,
                        size=9,
                        muted_color=color_mapper,
                        muted_alpha=0.2,
                        source=source,
                        legend=legend,
                    )

        color_bar = ColorBar(
            color_mapper=color_mapper["transform"],
            width=8,
            location=(0, 0),
            title="log dec",
            title_text_font_style="bold italic",
            title_text_align="center",
            major_label_text_align="left",
        )
        camp.add_tools(hover)
        camp.legend.background_fill_alpha = 0.1
        camp.legend.click_policy = "mute"
        camp.legend.location = "top_left"
        camp.add_layout(color_bar, "right")

        return camp

    def plot(self, *args, plot_type="matplotlib", **kwargs):
        """Plot campbell results.
        Parameters
        ----------
        harmonics: list, optional
            List withe the harmonics to be plotted.
            The default is to plot 1x.
        plot_type: str
            Matplotlib or bokeh.
            The default is matplotlib
        fig : matplotlib figure, optional
            Figure to insert axes with log_dec colorbar.
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        Returns
        -------
        ax : matplotlib axes
            Returns the axes object with the plot.
        """
        if plot_type == "matplotlib":
            return self._plot_matplotlib(*args, **kwargs)
        elif plot_type == "bokeh":
            return self._plot_bokeh(*args, **kwargs)
        else:
            raise ValueError(f"")


class FrequencyResponseResults:
    def __init__(self, freq_resp, speed_range, magnitude, phase):
        self.freq_resp = freq_resp
        self.speed_range = speed_range
        self.magnitude = magnitude
        self.phase = phase

    def plot_magnitude_matplotlib(self, inp, out, ax=None, units="mic-pk-pk", **kwargs):
        """Plot frequency response.
        This method plots the frequency response magnitude given an output and
        an input.
        Parameters
        ----------
        inp : int
            Input.
        out : int
            Output.
        ax : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax : matplotlib.axes
            Matplotlib axes with amplitude plot.
        mag_plot : bokeh plot axes
            Bokeh plot axes with amplitude plot.
        Examples
        --------
        """
        if ax is None:
            ax = plt.gca()

        frequency_range = self.speed_range
        mag = self.magnitude

        ax.plot(frequency_range, mag[inp, out, :], **kwargs)

        ax.set_xlim(0, max(frequency_range))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="lower"))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="upper"))

        ax.set_ylabel("Mag H$(j\omega)$")
        ax.set_xlabel("Frequency (rad/s)")
        ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))

        return ax

    def plot_magnitude_bokeh(self, inp, out, units="m", **kwargs):
        """Plot frequency response.
        This method plots the frequency response magnitude given an output and
        an input.
        Parameters
        ----------
        inp : int
            Input.
        out : int
            Output.
        ax : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax : matplotlib.axes
            Matplotlib axes with amplitude plot.
        mag_plot : bokeh plot axes
            Bokeh plot axes with amplitude plot.
        Examples
        --------
        """
        frequency_range = self.speed_range
        mag = self.magnitude

        if units == "m":
            y_axis_label = "Amplitude (m)"
        elif units == "mic-pk-pk":
            y_axis_label = "Amplitude (\mu pk-pk)"
        else:
            y_axis_label = "Amplitude (dB)"

        # bokeh plot - create a new plot
        mag_plot = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            width=900,
            height=400,
            title="Frequency Response - Magnitude",
            x_axis_label="Frequency",
            y_axis_label=y_axis_label,
        )
        mag_plot.xaxis.axis_label_text_font_size = "14pt"
        mag_plot.yaxis.axis_label_text_font_size = "14pt"

        source = ColumnDataSource(dict(x=frequency_range, y=mag[inp, out, :]))
        mag_plot.line(
            x="x",
            y="y",
            source=source,
            line_color=bokeh_colors[0],
            line_alpha=1.0,
            line_width=3,
        )

        return mag_plot

    def plot_phase_matplotlib(self, inp, out, ax=None, **kwargs):
        """Plot frequency response.
        This method plots the frequency response phase given an output and
        an input.
        Parameters
        ----------
        inp : int
            Input.
        out : int
            Output.
        ax : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax : matplotlib.axes
            Matplotlib axes with phase plot.
        phase_plot : bokeh plot axes
            Bokeh plot axes with phase plot.
        Examples
        --------
        """
        if ax is None:
            ax = plt.gca()

        frequency_range = self.speed_range
        phase = self.phase

        ax.plot(frequency_range, phase[inp, out, :], **kwargs)

        ax.set_xlim(0, max(frequency_range))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="lower"))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="upper"))

        ax.set_ylabel("Phase")
        ax.set_xlabel("Frequency (rad/s)")

        return ax

    def plot_phase_bokeh(self, inp, out, **kwargs):
        """Plot frequency response.
        This method plots the frequency response phase given an output and
        an input.
        Parameters
        ----------
        inp : int
            Input.
        out : int
            Output.
        ax : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax : matplotlib.axes
            Matplotlib axes with phase plot.
        phase_plot : bokeh plot axes
            Bokeh plot axes with phase plot.
        Examples
        --------
        """
        frequency_range = self.speed_range
        phase = self.phase

        # bokeh plot - create a new plot
        phase_plot = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            width=900,
            height=400,
            title="Frequency Response - Phase",
            x_axis_label="Frequency",
            y_axis_label="Phase",
        )
        phase_plot.xaxis.axis_label_text_font_size = "14pt"
        phase_plot.yaxis.axis_label_text_font_size = "14pt"

        source = ColumnDataSource(dict(x=frequency_range, y=phase[inp, out, :]))
        phase_plot.line(
            x="x",
            y="y",
            source=source,
            line_color=bokeh_colors[0],
            line_alpha=1.0,
            line_width=3,
        )

        return phase_plot

    def _plot_matplotlib(self, inp, out, ax0=None, ax1=None, **kwargs):
        """Plot frequency response.
        This method plots the frequency response given
        an output and an input.
        Parameters
        ----------
        inp : int
            Input.
        out : int
            Output.
        ax0 : matplotlib.axes, bokeh plot axes optional
            Matplotlib and bokeh plot axes where the amplitude will be plotted.
            If None creates a new.
        ax1 : matplotlib.axes, bokeh plot axes optional
            Matplotlib and bokeh plot axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax0 : matplotlib.axes
            Matplotlib axes with amplitude plot.
        ax1 : matplotlib.axes
            Matplotlib axes with phase plot.
        bk_ax0 : bokeh plot axes
            Bokeh plot axes with amplitude plot
        bk_ax1 : bokeh plot axes
            Bokeh plot axes with phase plot
        Examples
        --------
        """
        if ax0 is None and ax1 is None:
            fig, (ax0, ax1) = plt.subplots(2)

        # matplotlib axes
        ax0 = self.plot_magnitude_matplotlib(inp, out, ax=ax0)
        ax1 = self.plot_phase_matplotlib(inp, out, ax=ax1)

        ax0.set_xlabel("")

        return ax0, ax1

    def _plot_bokeh(self, inp, out, ax0=None, ax1=None, **kwargs):
        """Plot frequency response.
        This method plots the frequency response given
        an output and an input.
        Parameters
        ----------
        inp : int
            Input.
        out : int
            Output.
        ax0 : matplotlib.axes, bokeh plot axes optional
            Matplotlib and bokeh plot axes where the amplitude will be plotted.
            If None creates a new.
        ax1 : matplotlib.axes, bokeh plot axes optional
            Matplotlib and bokeh plot axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax0 : matplotlib.axes
            Matplotlib axes with amplitude plot.
        ax1 : matplotlib.axes
            Matplotlib axes with phase plot.
        bk_ax0 : bokeh plot axes
            Bokeh plot axes with amplitude plot
        bk_ax1 : bokeh plot axes
            Bokeh plot axes with phase plot
        Examples
        --------
        """
        # bokeh plot axes
        bk_ax0 = self.plot_magnitude_bokeh(inp, out, ax=ax0)
        bk_ax1 = self.plot_phase_bokeh(inp, out, ax=ax1)

        # show the bokeh plot results
        grid_plots = gridplot([[bk_ax0], [bk_ax1]])
        show(grid_plots)

        return bk_ax0, bk_ax1

    def plot(self, inp, out, *args, plot_type="matplotlib", **kwargs):
        """Plot frequency response.
        This method plots the frequency response given
        an output and an input.
        Parameters
        ----------
        inp : int
            Input.
        out : int
            Output.
        plot_type: str
            Matplotlib or bokeh.
            The default is matplotlib
        ax0 : matplotlib.axes, bokeh plot axes optional
            Matplotlib and bokeh plot axes where the amplitude will be plotted.
            If None creates a new.
        ax1 : matplotlib.axes, bokeh plot axes optional
            Matplotlib and bokeh plot axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax0 : matplotlib.axes
            Matplotlib axes with amplitude plot.
        ax1 : matplotlib.axes
            Matplotlib axes with phase plot.
        bk_ax0 : bokeh plot axes
            Bokeh plot axes with amplitude plot
        bk_ax1 : bokeh plot axes
            Bokeh plot axes with phase plot
        Examples
        --------
        """
        if plot_type == "matplotlib":
            return self._plot_matplotlib(inp, out, *args, **kwargs)
        elif plot_type == "bokeh":
            return self._plot_bokeh(inp, out, *args, **kwargs)
        else:
            raise ValueError(f"")

    def plot_freq_response_grid(self, outs, inps, ax=None, **kwargs):
        """Plot frequency response.
        This method plots the frequency response given
        an output and an input.
        Parameters
        ----------
        outs : list
            List with the desired outputs.
        inps : list
            List with the desired outputs.
        ax : array with matplotlib.axes, optional
            Matplotlib axes array created with plt.subplots.
            It needs to have a shape of (2*inputs, outputs).
        Returns
        -------
        ax : array with matplotlib.axes, optional
            Matplotlib axes array created with plt.subplots.
        """
        if ax is None:
            fig, ax = plt.subplots(
                len(inps) * 2,
                len(outs),
                sharex=True,
                figsize=(4 * len(outs), 3 * len(inps)),
            )
            fig.subplots_adjust(hspace=0.001, wspace=0.25)

        if len(outs) > 1:
            for i, out in enumerate(outs):
                for j, inp in enumerate(inps):
                    self.plot_magnitude(out, inp, ax=ax[2 * i, j], **kwargs)
                    self.plot_phase(out, inp, ax=ax[2 * i + 1, j], **kwargs)
        else:
            for i, inp in enumerate(inps):
                self.plot_magnitude(outs[0], inp, ax=ax[2 * i], **kwargs)
                self.plot_phase(outs[0], inp, ax=ax[2 * i + 1], **kwargs)

        return ax


class ForcedResponseResults:
    def __init__(self, forced_resp, speed_range, magnitude, phase):
        self.forced_resp = forced_resp
        self.speed_range = speed_range
        self.magnitude = magnitude
        self.phase = phase

    def plot_magnitude_matplotlib(self, dof, ax=None, units="m", **kwargs):
        """Plot frequency response.
        This method plots the frequency response magnitude given an output and
        an input.

        Parameters
        ----------
        dof : int
            Degree of freedom.
        ax : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        units : str
            Units to plot the magnitude ('m' or 'mic-pk-pk')
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')

        Returns
        -------
        ax : matplotlib.axes
            Matplotlib axes with phase plot.

        Examples
        --------
        """
        if ax is None:
            ax = plt.gca()

        frequency_range = self.speed_range
        mag = self.magnitude

        if units == "m":
            ax.set_ylabel("Amplitude $(m)$")
        elif units == "mic-pk-pk":
            mag = 2 * mag * 1e6
            ax.set_ylabel("Amplitude $(\mu pk-pk)$")

        ax.plot(frequency_range, mag[dof], **kwargs)

        ax.set_xlim(0, max(frequency_range))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="lower"))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="upper"))

        ax.set_xlabel("Frequency (rad/s)")
        ax.legend()

        return ax

    def plot_magnitude_bokeh(self, dof, units="m", **kwargs):
        """Plot frequency response.
        This method plots the frequency response magnitude given an output and
        an input.

        Parameters
        ----------
        dof : int
            Degree of freedom.
        units : str
            Units to plot the magnitude ('m' or 'mic-pk-pk')
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')

        Returns
        -------
        mag_plot : bokeh axes
            bokeh axes with magnitude plot

        Examples
        --------
        """
        frequency_range = self.speed_range
        mag = self.magnitude

        if units == "m":
            y_axis_label = "Amplitude (m)"
        elif units == "mic-pk-pk":
            mag = 2 * mag * 1e6
            y_axis_label = "Amplitude $(\mu pk-pk)$"

        # bokeh plot - create a new plot
        mag_plot = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            width=900,
            height=400,
            title="Forced Response - Magnitude",
            x_axis_label="Frequency",
            x_range=[0, max(frequency_range)],
            y_axis_label=y_axis_label,
        )
        mag_plot.xaxis.axis_label_text_font_size = "14pt"
        mag_plot.yaxis.axis_label_text_font_size = "14pt"

        source = ColumnDataSource(dict(x=frequency_range, y=mag[dof]))
        mag_plot.line(
            x="x",
            y="y",
            source=source,
            line_color=bokeh_colors[0],
            line_alpha=1.0,
            line_width=3,
        )

        return mag_plot

    def plot_phase_matplotlib(self, dof, ax=None, **kwargs):
        """Plot frequency response.
        This method plots the frequency response phase given an output and
        an input.

        Parameters
        ----------
        dof : int
            Degree of freedom.
        ax : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')

        Returns
        -------
        ax : matplotlib.axes
            Matplotlib axes with phase plot.

        Examples
        --------
        """
        if ax is None:
            ax = plt.gca()

        frequency_range = self.speed_range
        phase = self.phase

        ax.plot(frequency_range, phase[dof], **kwargs)

        ax.set_xlim(0, max(frequency_range))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="lower"))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="upper"))

        ax.set_ylabel("Phase")
        ax.set_xlabel("Frequency (rad/s)")
        ax.legend()

        return ax

    def plot_phase_bokeh(self, dof, **kwargs):
        """Plot frequency response.
        This method plots the frequency response phase given an output and
        an input.

        Parameters
        ----------
        dof : int
            Degree of freedom.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')

        Returns
        -------
        phase_plot : bokeh axes
            Bokeh axes with phase plot

        Examples
        --------
        """
        frequency_range = self.speed_range
        phase = self.phase

        phase_plot = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            width=900,
            height=400,
            title="Forced Response - Magnitude",
            x_axis_label="Frequency",
            x_range=[0, max(frequency_range)],
            y_axis_label="Phase",
        )
        source = ColumnDataSource(dict(x=frequency_range, y=phase[dof]))
        phase_plot.line(
            x="x",
            y="y",
            source=source,
            line_color=bokeh_colors[0],
            line_alpha=1.0,
            line_width=3,
        )
        phase_plot.xaxis.axis_label_text_font_size = "14pt"
        phase_plot.yaxis.axis_label_text_font_size = "14pt"

        return phase_plot

    def _plot_matplotlib(self, dof, ax0=None, ax1=None, **kwargs):

        if ax0 is None and ax1 is None:
            fig, (ax0, ax1) = plt.subplots(2)

        ax0 = self.plot_magnitude_matplotlib(dof, ax=ax0, **kwargs)
        # remove label from phase plot
        kwargs.pop("label", None)
        kwargs.pop("units", None)
        ax1 = self.plot_phase_matplotlib(dof, ax=ax1, **kwargs)

        ax0.set_xlabel("")
        ax0.legend()

        return ax0, ax1

    def _plot_bokeh(self, dof, **kwargs):
        # bokeh plot axes
        bk_ax0 = self.plot_magnitude_bokeh(dof, **kwargs)
        bk_ax1 = self.plot_phase_bokeh(dof, **kwargs)

        # show the bokeh plot results
        grid_plots = gridplot([[bk_ax0], [bk_ax1]])
        show(grid_plots)

        return bk_ax0, bk_ax1

    def plot(self, dof, plot_type="matplotlib", **kwargs):
        """Plot frequency response.
        This method plots the frequency response given
        an output and an input.

        Parameters
        ----------
        dof : int
            Degree of freedom.
        plot_type: str
            Matplotlib or bokeh.
            The default is matplotlib
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')

        Returns
        -------

        Examples
        --------
        """
        if plot_type == "matplotlib":
            return self._plot_matplotlib(dof, **kwargs)
        elif plot_type == "bokeh":
            return self._plot_bokeh(dof, **kwargs)
        else:
            raise ValueError(f"{plot_type} is not a valid plot type.")


class ModeShapeResults:
    def __init__(
        self,
        modes,
        ndof,
        nodes,
        nodes_pos,
        elements_length,
        w,
        wd,
        log_dec,
        kappa_modes,
    ):
        self.modes = modes
        self.ndof = ndof
        self.nodes = nodes
        self.nodes_pos = nodes_pos
        self.elements_length = elements_length
        self.w = w
        self.wd = wd
        self.log_dec = log_dec
        self.kappa_modes = kappa_modes

    def plot(self, mode=None, evec=None, fig=None, ax=None):
        if ax is None:
            fig = plt.figure()
            ax = fig.gca(projection="3d")

        evec0 = self.modes[:, mode]
        nodes = self.nodes
        nodes_pos = self.nodes_pos
        kappa_modes = self.kappa_modes
        elements_length = self.elements_length

        modex = evec0[0::4]
        modey = evec0[1::4]
        xmax, ixmax = max(abs(modex)), np.argmax(abs(modex))
        ymax, iymax = max(abs(modey)), np.argmax(abs(modey))

        if ymax > 0.4 * xmax:
            evec0 /= modey[iymax]
        else:
            evec0 /= modex[ixmax]

        modex = evec0[0::4]
        modey = evec0[1::4]

        num_points = 201
        c = np.linspace(0, 2 * np.pi, num_points)
        circle = np.exp(1j * c)

        x_circles = np.zeros((num_points, len(nodes)))
        y_circles = np.zeros((num_points, len(nodes)))
        z_circles_pos = np.zeros((num_points, len(nodes)))

        kappa_mode = kappa_modes[mode]

        for node in nodes:
            x = modex[node] * circle
            x_circles[:, node] = np.real(x)
            y = modey[node] * circle
            y_circles[:, node] = np.real(y)
            z_circles_pos[:, node] = nodes_pos[node]

        # plot lines
        nn = 21
        zeta = np.linspace(0, 1, nn)
        onn = np.ones_like(zeta)

        zeta = zeta.reshape(nn, 1)
        onn = onn.reshape(nn, 1)

        xn = np.zeros(nn * (len(nodes) - 1))
        yn = np.zeros(nn * (len(nodes) - 1))
        zn = np.zeros(nn * (len(nodes) - 1))

        N1 = onn - 3 * zeta ** 2 + 2 * zeta ** 3
        N2 = zeta - 2 * zeta ** 2 + zeta ** 3
        N3 = 3 * zeta ** 2 - 2 * zeta ** 3
        N4 = -zeta ** 2 + zeta ** 3

        for Le, n in zip(elements_length, nodes):
            node_pos = nodes_pos[n]
            Nx = np.hstack((N1, Le * N2, N3, Le * N4))
            Ny = np.hstack((N1, -Le * N2, N3, -Le * N4))

            xx = [4 * n, 4 * n + 3, 4 * n + 4, 4 * n + 7]
            yy = [4 * n + 1, 4 * n + 2, 4 * n + 5, 4 * n + 6]

            pos0 = nn * n
            pos1 = nn * (n + 1)
            xn[pos0:pos1] = Nx @ evec0[xx].real
            yn[pos0:pos1] = Ny @ evec0[yy].real
            zn[pos0:pos1] = (node_pos * onn + Le * zeta).reshape(nn)

        for node in nodes:
            ax.plot(
                x_circles[10:, node],
                y_circles[10:, node],
                z_circles_pos[10:, node],
                color=kappa_mode[node],
                linewidth=0.5,
                zdir="x",
            )
            ax.scatter(
                x_circles[10, node],
                y_circles[10, node],
                z_circles_pos[10, node],
                s=5,
                color=kappa_mode[node],
                zdir="x",
            )

        ax.plot(xn, yn, zn, "k--", zdir="x")

        # plot center line
        zn_cl0 = -(zn[-1] * 0.1)
        zn_cl1 = zn[-1] * 1.1
        zn_cl = np.linspace(zn_cl0, zn_cl1, 30)
        ax.plot(zn_cl * 0, zn_cl * 0, zn_cl, "k-.", linewidth=0.8, zdir="x")

        ax.set_zlim(-2, 2)
        ax.set_ylim(-2, 2)
        ax.set_xlim(zn_cl0 - 0.1, zn_cl1 + 0.1)

        ax.set_title(
            f"$speed$ = {self.w:.1f} rad/s\n$"
            f"\omega_d$ = {self.wd[mode]:.1f} rad/s\n"
            f"$log dec$ = {self.log_dec[mode]:.1f}"
        )

        return fig, ax


class StaticResults():
    def __init__(
        self,
        disp_y,
        Vx,
        Bm,
        df_shaft,
        df_disks,
        df_bearings,
        nodes,
        nodes_pos,
        Vx_axis,
    ):

        self.disp_y = disp_y
        self.Vx = Vx
        self.Bm = Bm
        self.df_shaft = df_shaft
        self.df_disks = df_disks
        self.df_bearings = df_bearings
        self.nodes = nodes
        self.nodes_pos = nodes_pos
        self.Vx_axis = Vx_axis

    def plot(self):
        """Plot static analysis graphs.
        This method plots:
            free-body diagram,
            deformed shaft,
            shearing force diagram,
            bending moment diagram.

        Parameters
        ----------

        Returns
        -------
        grid_plots : bokeh.gridplot
        --------
        """
        source = ColumnDataSource(
            data=dict(
                x0=self.nodes_pos,
                y0=self.disp_y * 1000,
                y1=[0] * len(self.nodes_pos)
            )
        )

        TOOLS = "pan,wheel_zoom,box_zoom,reset,save,box_select,hover"
        TOOLTIPS = [
            ("Shaft lenght:", "@x0"),
            ("Underformed:", "@y1"),
            ("Displacement:", "@y0"),
        ]

        # create displacement plot
        disp_graph = figure(
            tools=TOOLS,
            tooltips=TOOLTIPS,
            width=800,
            height=400,
            title="Static Analysis",
            x_axis_label="shaft lenght",
            y_axis_label="lateral displacement",
        )
        disp_graph.xaxis.axis_label_text_font_size = "14pt"
        disp_graph.yaxis.axis_label_text_font_size = "14pt"

        interpolated = interpolate.interp1d(
            source.data["x0"], source.data["y0"], kind="cubic"
        )
        xnew = np.linspace(
            source.data["x0"][0],
            source.data["x0"][-1],
            num=len(self.nodes_pos) * 20,
            endpoint=True,
        )

        ynew = interpolated(xnew)
        auxsource = ColumnDataSource(data=dict(x0=xnew, y0=ynew, y1=[0] * len(xnew)))

        disp_graph.line(
            "x0",
            "y0",
            source=auxsource,
            legend="Deformed shaft",
            line_width=3,
            line_color=bokeh_colors[9],
        )
        disp_graph.circle(
            "x0",
            "y0",
            source=source,
            legend="Deformed shaft",
            size=8,
            fill_color=bokeh_colors[9],
        )
        disp_graph.line(
            "x0",
            "y1",
            source=source,
            legend="underformed shaft",
            line_width=3,
            line_color=bokeh_colors[0],
        )
        disp_graph.circle(
            "x0",
            "y1",
            source=source,
            legend="underformed shaft",
            size=8,
            fill_color=bokeh_colors[0],
        )

        # create a new plot for free body diagram (FDB)
        y_range = []
        sh_weight = sum(self.df_shaft["m"].values) * 9.8065
        y_range.append(sh_weight)
        for i, node in enumerate(self.df_bearings["n"]):
            y_range.append(-self.disp_y[node] * self.df_bearings.loc[i, "kyy"].coefficient[0])

        shaft_end = self.nodes_pos[-1]
        FBD = figure(
            tools=TOOLS,
            width=800,
            height=400,
            title="Free-Body Diagram",
            x_axis_label="shaft lenght",
            y_axis_label="Force",
            x_range=[-0.1 * shaft_end, 1.1 * shaft_end],
            y_range=[-max(y_range) * 1.4, max(y_range) * 1.4],
        )
        FBD.xaxis.axis_label_text_font_size = "14pt"
        FBD.yaxis.axis_label_text_font_size = "14pt"

        FBD.line("x0", "y1", source=source, line_width=5, line_color=bokeh_colors[0])

        # FBD - plot arrows indicating shaft weight distribution
        text = str("%.1f" % sh_weight)
        FBD.line(
            x=self.nodes_pos,
            y=[sh_weight] * len(self.nodes_pos),
            line_width=2,
            line_color=bokeh_colors[0],
        )

        for node in self.nodes_pos:
            FBD.add_layout(
                Arrow(
                    end=NormalHead(
                        fill_color=bokeh_colors[7],
                        fill_alpha=1.0,
                        size=16,
                        line_width=2,
                        line_color=bokeh_colors[0],
                    ),
                    x_start=node,
                    y_start=sh_weight,
                    x_end=node,
                    y_end=0,
                )
            )

        FBD.add_layout(
            Label(
                x=self.nodes_pos[0],
                y=sh_weight,
                text="W = " + text + "N",
                text_font_style="bold",
                text_baseline="top",
                text_align="left",
                y_offset=20,
            )
        )

        # FBD - calculate the reaction force of bearings and plot arrows
        for i, node in enumerate(self.df_bearings["n"]):
            Fb = -self.disp_y[node] * self.df_bearings.loc[i, "kyy"].coefficient[0]
            text = str("%.1f" % Fb)
            FBD.add_layout(
                Arrow(
                    end=NormalHead(
                        fill_color=bokeh_colors[7],
                        fill_alpha=1.0,
                        size=16,
                        line_width=2,
                        line_color=bokeh_colors[0],
                    ),
                    x_start=self.nodes_pos[node],
                    y_start=-Fb,
                    x_end=self.nodes_pos[node],
                    y_end=0,
                )
            )
            FBD.add_layout(
                Label(
                    x=self.nodes_pos[node],
                    y=-Fb,
                    text="Fb = " + text + "N",
                    text_font_style="bold",
                    text_baseline="bottom",
                    text_align="center",
                    y_offset=-20,
                )
            )

        # FBD - plot arrows indicating disk weight
        if len(self.df_disks) != 0:
            for i, node in enumerate(self.df_disks["n"]):
                Fd = self.df_disks.loc[i, "m"] * 9.8065
                text = str("%.1f" % Fd)
                FBD.add_layout(
                    Arrow(
                        end=NormalHead(
                            fill_color=bokeh_colors[7],
                            fill_alpha=1.0,
                            size=16,
                            line_width=2,
                            line_color=bokeh_colors[0],
                        ),
                        x_start=self.nodes_pos[node],
                        y_start=Fd,
                        x_end=self.nodes_pos[node],
                        y_end=0,
                    )
                )
                FBD.add_layout(
                    Label(
                        x=self.nodes_pos[node],
                        y=Fd,
                        text="Fd = " + text + "N",
                        text_font_style="bold",
                        text_baseline="top",
                        text_align="center",
                        y_offset=20,
                    )
                )

        # Shearing Force Diagram plot (SF)
        source_SF = ColumnDataSource(data=dict(x=self.Vx_axis, y=self.Vx))
        TOOLTIPS_SF = [("Shearing Force:", "@y")]
        SF = figure(
            tools=TOOLS,
            tooltips=TOOLTIPS_SF,
            width=800,
            height=400,
            title="Shearing Force Diagram",
            x_axis_label="Shaft lenght",
            y_axis_label="Force",
            x_range=[-0.1 * shaft_end, 1.1 * shaft_end],
        )
        SF.xaxis.axis_label_text_font_size = "14pt"
        SF.yaxis.axis_label_text_font_size = "14pt"

        SF.line("x", "y", source=source_SF, line_width=4, line_color=bokeh_colors[0])
        SF.circle("x", "y", source=source_SF, size=8, fill_color=bokeh_colors[0])

        # SF - plot centerline
        SF.line(
            [-0.1 * shaft_end, 1.1 * shaft_end],
            [0, 0],
            line_width=3,
            line_dash="dotdash",
            line_color=bokeh_colors[0],
        )

        # Bending Moment Diagram plot (BM)
        source_BM = ColumnDataSource(data=dict(x=self.nodes_pos, y=self.Bm))
        TOOLTIPS_BM = [("Bending Moment:", "@y")]
        BM = figure(
            tools=TOOLS,
            tooltips=TOOLTIPS_BM,
            width=800,
            height=400,
            title="Bending Moment Diagram",
            x_axis_label="Shaft lenght",
            y_axis_label="Bending Moment",
            x_range=[-0.1 * shaft_end, 1.1 * shaft_end],
        )
        BM.xaxis.axis_label_text_font_size = "14pt"
        BM.yaxis.axis_label_text_font_size = "14pt"

        i = 0
        while True:
            if i + 3 > len(self.nodes):
                break

            interpolated_BM = interpolate.interp1d(
                self.nodes_pos[i : i + 3], self.Bm[i : i + 3], kind="quadratic"
            )
            xnew_BM = np.linspace(self.nodes_pos[i], self.nodes_pos[i + 2], num=42, endpoint=True)

            ynew_BM = interpolated_BM(xnew_BM)
            auxsource_BM = ColumnDataSource(data=dict(x=xnew_BM, y=ynew_BM))
            BM.line(
                "x", "y", source=auxsource_BM, line_width=4, line_color=bokeh_colors[0]
            )
            i += 2
        BM.circle("x", "y", source=source_BM, size=8, fill_color=bokeh_colors[0])

        # BM - plot centerline
        BM.line(
            [-0.1 * shaft_end, 1.1 * shaft_end],
            [0, 0],
            line_width=3,
            line_dash="dotdash",
            line_color=bokeh_colors[0],
        )

        grid_plots = gridplot([[FBD, SF], [disp_graph, BM]])

        show(grid_plots)


class ConvergenceResults:
    def __init__(self, el_num, eigv_arr, error_arr):
        self.el_num = el_num
        self.eigv_arr = eigv_arr
        self.error_arr = error_arr

    def plot(self):
        """This method plots:
            Natural Frequency vs Number of Elements
            Relative Error vs Number of Elements

        Parameters
        ----------

        Returns
        -------
        plot : bokeh.figure
            Bokeh plot showing the results
        --------
        """
        source = ColumnDataSource(
            data=dict(x0=self.el_num, y0=self.eigv_arr, y1=self.error_arr)
        )

        TOOLS = "pan,wheel_zoom,box_zoom,hover,reset,save,"
        TOOLTIPS1 = [("Frequency:", "@y0"), ("Number of Elements", "@x0")]
        TOOLTIPS2 = [("Relative Error:", "@y1"), ("Number of Elements", "@x0")]

        # create a new plot and add a renderer
        freq_arr = figure(
            tools=TOOLS,
            tooltips=TOOLTIPS1,
            width=800,
            height=600,
            title="Frequency Evaluation",
            x_axis_label="Numer of Elements",
            y_axis_label="Frequency (rad/s)",
        )
        freq_arr.xaxis.axis_label_text_font_size = "14pt"
        freq_arr.yaxis.axis_label_text_font_size = "14pt"

        freq_arr.line("x0", "y0", source=source, line_width=3, line_color="crimson")
        freq_arr.circle("x0", "y0", source=source, size=8, fill_color="crimson")

        # create another new plot and add a renderer
        rel_error = figure(
            tools=TOOLS,
            tooltips=TOOLTIPS2,
            width=800,
            height=600,
            title="Relative Error Evaluation",
            x_axis_label="Number of Elements",
            y_axis_label="Relative Error (%)",
        )
        rel_error.xaxis.axis_label_text_font_size = "14pt"
        rel_error.yaxis.axis_label_text_font_size = "14pt"

        rel_error.line(
            "x0", "y1", source=source, line_width=3, line_color="darkslategray"
        )
        rel_error.circle("x0", "y1", source=source, fill_color="darkslategray", size=8)

        # put the subplots in a gridplot
        plot = gridplot([[freq_arr, rel_error]])
        # show the plots
        show(plot)

        return plot


class TimeResponseResults:
    def __init__(self, t, yout, xout, dof):
        self.t = t
        self.yout = yout
        self.xout = xout
        self.dof = dof

    def _plot_matplotlib(self, ax=None):

        if ax is None:
            ax = plt.gca()

        ax.plot(self.t, self.yout[:, self.dof])

        if self.dof % 4 == 0:
            obs_dof = "x"
            amp = "m"
        elif self.dof % 4 == 1:
            obs_dof = "y"
            amp = "m"
        elif self.dof % 4 == 2:
            obs_dof = "\u03B1"  # unicode for alpha
            amp = "rad"
        elif self.dof % 4 == 3:
            obs_dof = "\u03B2"  # unicode for beta
            amp = "rad"

        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Amplitude (%s)" % amp)
        ax.set_title(
            "Response for node %s and degree of freedom %s" % (self.dof // 4, obs_dof)
        )

    def _plot_bokeh(self):

        if self.dof % 4 == 0:
            obs_dof = "x"
            amp = "m"
        elif self.dof % 4 == 1:
            obs_dof = "y"
            amp = "m"
        elif self.dof % 4 == 2:
            obs_dof = "\u03B1"  # unicode for alpha
            amp = "rad"
        elif self.dof % 4 == 3:
            obs_dof = "\u03B2"  # unicode for beta
            amp = "rad"

        # bokeh plot - create a new plot
        bk_ax = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            width=1200,
            height=900,
            title="Response for node %s and degree of freedom %s" % (self.dof // 4, obs_dof),
            x_axis_label="Time (s)",
            y_axis_label="Amplitude (%s)" % amp,
        )
        bk_ax.xaxis.axis_label_text_font_size = "14pt"
        bk_ax.yaxis.axis_label_text_font_size = "14pt"

        # bokeh plot - plot shaft centerline
        bk_ax.line(
                self.t,
                self.yout[:, self.dof],
                line_width=3,
                line_color=bokeh_colors[0]
        )

        show(bk_ax)

    def plot(self, plot_type="matplotlib", **kwargs):
        """Plot time response.

        This function will take a rotor object and plot its time response
        given a force and a time.

        Parameters
        ----------
        dof : int
            Degree of freedom.
        plot_type: str
            Matplotlib or bokeh.
            The default is matplotlib
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')

        Returns
        -------

        Examples
        --------
        """
        if plot_type == "matplotlib":
            return self._plot_matplotlib(**kwargs)
        elif plot_type == "bokeh":
            return self._plot_bokeh(**kwargs)
        else:
            raise ValueError(f"{plot_type} is not a valid plot type.")



"""Materials module.

This module defines the Material class and defines
some of the most common materials used in rotors.
"""
import os

import toml

import ross as rs

__all__ = ["Material", "steel"]


class Material:
    """Material.

    Class used to create a material and define its properties.
    Density and at least 2 arguments from E, G_s and Poisson
    should be provided.

    Run Material.available_materials() for materials already provided.

    Parameters
    ----------
    name : str
        Material name.
    E : float
        Young's modulus (N/m**2).
    G_s : float
        Shear modulus (N/m**2).
    rho : float
        Density (N/m**3).
    color : str
        Can be used on plots.

    Examples
    --------
    >>> AISI4140 = Material(name='AISI4140', rho=7850, E=203.2e9, G_s=80e9)
    >>> Steel = Material(name="Steel", rho=7810, E=211e9, G_s=81.2e9)
    >>> AISI4140.Poisson
    0.27

    """

    def __init__(self, name, rho, **kwargs):

        assert name is not None, "Name not provided"
        assert type(name) is str, "Name must be a string"
        assert " " not in name, "Spaces are not allowed in Material name"
        assert (
            sum([1 if i in ["E", "G_s", "Poisson"] else 0 for i in kwargs]) > 1
        ), "At least 2 arguments from E, G_s and Poisson should be provided"

        self.name = name
        self.rho = rho
        self.E = kwargs.get("E", None)
        self.Poisson = kwargs.get("Poisson", None)
        self.G_s = kwargs.get("G_s", None)
        self.color = kwargs.get("color", "#525252")

        if self.E is None:
            self.E = self.G_s * (2 * (1 + self.Poisson))
        elif self.G_s is None:
            self.G_s = self.E / (2 * (1 + self.Poisson))
        elif self.Poisson is None:
            self.Poisson = (self.E / (2 * self.G_s)) - 1

    def __eq__(self, other):
        if self.__dict__ == other.__dict__:
            return True
        else:
            return False

    def __repr__(self):
        selfE = "{:.2e}".format(self.E)
        selfPoisson = "{:.2e}".format(self.Poisson)
        selfrho = "{:.2e}".format(self.rho)
        selfGs = "{:.2e}".format(self.G_s)

        return (
            f"Material"
            f"(name={self.name}, rho={selfrho}, G_s={selfGs} "
            f"E={selfE}, Poisson={selfPoisson} color={self.color!r}"
        )

    def __str__(self):
        return (
            f"{self.name}"
            f'\n{35*"-"}'
            f"\nDensity         (N/m**3): {float(self.rho):{2}.{8}}"
            f"\nYoung`s modulus (N/m**2): {float(self.E):{2}.{8}}"
            f"\nShear modulus   (N/m**2): {float(self.G_s):{2}.{8}}"
            f"\nPoisson coefficient     : {float(self.Poisson):{2}.{8}}"
        )

    @staticmethod
    def dump_data(data):
        with open("available_materials.toml", "w") as f:
            toml.dump(data, f)

    @staticmethod
    def load_data():
        try:
            with open("available_materials.toml", "r") as f:
                data = toml.load(f)
        except FileNotFoundError:
            data = {"Materials": {}}
            Material.dump_data(data)
        return data

    @staticmethod
    def use_material(name):
        run_path = os.getcwd()
        ross_path = os.path.dirname(rs.__file__)
        os.chdir(ross_path)

        data = Material.load_data()
        try:
            material = data["Materials"][name]
            return Material(**material)
        except KeyError:
            raise KeyError("There isn't a instanced material with this name.")
        os.chdir(run_path)

    @staticmethod
    def remove_material(name):
        run_path = os.getcwd()
        ross_path = os.path.dirname(rs.__file__)
        os.chdir(ross_path)

        data = Material.load_data()
        try:
            del data["Materials"][name]
        except KeyError:
            return "There isn't a saved material with this name."
        Material.dump_data(data)
        os.chdir(run_path)

    @staticmethod
    def available_materials():
        run_path = os.getcwd()
        ross_path = os.path.dirname(rs.__file__)
        os.chdir(ross_path)

        try:
            data = Material.load_data()
            return list(data["Materials"].keys())
        except FileNotFoundError:
            return "There is no saved materials."
        os.chdir(run_path)

    def save_material(self):
        run_path = os.getcwd()
        ross_path = os.path.dirname(rs.__file__)
        os.chdir(ross_path)

        data = Material.load_data()
        data["Materials"][self.name] = self.__dict__
        Material.dump_data(data)
        os.chdir(run_path)

steel = Material(name="Steel", rho=7810, E=211e9, G_s=81.2e9)



from abc import ABC, abstractmethod
from collections import namedtuple

import pandas as pd
import toml


class Element(ABC):
    """Element class.
    This class is a general class to be called for other files which
    create specific elements for the user
    """

    def __init__(self, n):
        self.n = n
        self.dof_mapping = None
        pass

    def save(self, file_name):
        pass

    @staticmethod
    def load(self, file_name):
        pass

    @abstractmethod
    def M(self):
        """Mass matrix."""
        pass

    @abstractmethod
    def C(self, frequency):
        """Frequency dependent damping coefficients matrix."""
        pass

    @abstractmethod
    def K(self, frequency):
        """Frequency dependent stiffness coefficients matrix."""
        pass

    @abstractmethod
    def G(self):
        """Giroscopic matrix."""
        pass

    def summary(self):
        """A summary for the element.

        A pandas series with the element properties as variables.
        """
        attributes = self.__dict__
        attributes["type"] = self.__class__.__name__
        return pd.Series(attributes)

    @staticmethod
    def load_data(file_name):
        try:
            with open(file_name, "r") as f:
                data = toml.load(f)
                if data == {"": {}}:
                    data = {file_name[:-5]: {}}

        except FileNotFoundError:
            data = {file_name[:-5]: {}}
            Element.dump_data(data, file_name)
        return data

    @staticmethod
    def dump_data(data, file_name):
        with open(file_name, "w") as f:
            toml.dump(data, f)

    @abstractmethod
    def dof_mapping(self):
        """Should return a dictionary with a mapping between degree of freedom
        and its index.

        Example considering a shaft element:
        def dof_mapping(self):
            return dict(
                x0=0,
                y0=1,
                alpha0=2,
                beta0=3,
                x1=4,
                y1=5,
                alpha1=6,
                beta1=7,
            )
        """
        pass

    def dof_local_index(self):
        """Get the local index for a element specific degree of freedom."""
        dof_mapping = self.dof_mapping()
        dof_tuple = namedtuple("LocalIndex", dof_mapping)
        global_index = dof_tuple(**dof_mapping)

        return global_index

    def dof_global_index(self):
        """Get the global index for a element specific degree of freedom."""
        dof_mapping = self.dof_mapping()
        dof_tuple = namedtuple("GlobalIndex", dof_mapping)

        for k, v in dof_mapping.items():
            dof_mapping[k] = 4 * self.n + v

        global_index = dof_tuple(**dof_mapping)

        return global_index
